import jax.numpy as jnp
from jax import jit
import jax
from functools import partial
from jax.config import config

from jax.lax import cond, dynamic_update_slice, fori_loop, dynamic_slice
import diffrax

import h5py
import hdfdict

import scipy as scipy
import scipy.integrate
import scipy.optimize

import pkg_resources
from typing import Union

config.update("jax_enable_x64", True) # run in 64 bit by default or else you will lack the dynamic range required

class Box:
    """ Box class which represents each individual box in the carbon model."""

    def __init__(self, name, reservoir, production_coefficient=0.0, hemisphere='None'):
        """ init method to initialise object.
        Parameters
        ----------
        name : str
            name of the Box.
        reservoir : float
            reservoir content of the Box.
        production_coefficient : float, optional
            production coefficient of the Box. Defaults to 0.
        hemisphere : str, optional
            hemisphere that this box is in. Defaults to None (i.e. this box is not part of an inter-hemispheric system).
        """
        self._name = name
        self._reservoir = reservoir
        self._production = production_coefficient
        self._hemisphere = hemisphere
        assert reservoir >= 0
        assert production_coefficient >= 0
        assert hemisphere in ['None', 'north', 'south']

    def get_hemisphere(self):
        """ Getter method for the hemisphere of the Box Class.
        Returns
        -------
        str
            hemisphere the box is in.
        """
        return self._hemisphere

    def get_name(self):
        """ Getter method for the name of the Box Class.
        Returns
        -------
        str
            name of the Box.
        """
        return self._name

    def get_reservoir_content(self):
        """ Getter method for the reservoir content of the Box Class.
        Returns
        -------
        float
            reservoir content of the box.
        """
        return self._reservoir

    def get_production(self):
        """ Getter method for the production coefficient of the Box Class.
        Returns
        -------
         float
            production coefficient of the box.
        """
        return self._production

    def __str__(self):
        """ Overrides the default string behaviour to display a user-friendly output.
        Returns
        -------
        str
            string representation of the Box Object returned in the following form -
            name:reservoir size:production coefficient
        """
        return self._name + ":" + str(self._reservoir) + ":" + str(self._production)


class Flow:
    """ Flow class to imitate the fluxes between boxes in a carbon box model."""

    def __init__(self, source, destination, flux_rate):
        """ Init method to initialise object.
        Parameters
        ----------
        source : Box
            source of the flow.
        destination : Box
            destination the flow.
        flux_rate : float
            flux rate between the source and the destination. Flux must be a non-negative value.
        """
        self._source = source
        self._destination = destination
        assert flux_rate >= 0
        self._flux = flux_rate

    def get_source(self):
        """ Getter method for the source node of the Flow Class.
        Returns
        -------
        Box
            source of flow.
        """
        return self._source

    def get_destination(self):
        """ Getter method for the destination node of the Flow Class.
        Returns
        -------
        Box
            destination of the flow.
        """
        return self._destination

    def get_flux(self):
        """ Getter method for the flux rate of the Flow Class.
        Returns
        -------
        float
            flux of the flow object.
        """
        return self._flux

    def __str__(self):
        """ Overrides the default string behaviour to display a user-friendly output.
        Returns
        -------
        str
            string representation of the Flow Object returned in the following form -
            str(source)-->str(destination):flux_value
        """
        return str(self._source) + " --> " + str(self._destination) + " : " + str(self._flux)


class CarbonBoxModel:
    """Carbon Box Model class which represents the box model which is made up of Box Objects and Flux Objects."""

    def __init__(self, production_rate_units='kg/yr', flow_rate_units='Gt/yr'):
        """ Init method for creating an object of this class.
        Parameters
        ----------
        production_rate_units : str, optional
            units for the production rate. Only valid units are 'kg/yr' or 'atoms/cm^2/s'.
        flow_rate_units : str, optional
            units for the flow rate. Only valid values are 'Gt/yr' or '1/yr'.
        """
        self._non_hemisphere_model = False
        self._nodes = {}
        self._reverse_nodes = {}
        self._edges = []
        self._n_nodes = 0
        self._reservoir_content = None
        self._fluxes = None
        self._decay_matrix = None
        self._production_coefficients = None
        self._decay_constant = jnp.log(2) / 5700
        self._production_rate_units = production_rate_units
        self._flow_rate_units = flow_rate_units
        self._corrected_fluxes = None
        self._matrix = None

    def add_nodes(self, nodes):
        """ Adds the nodes to the Carbon Box Model. If the node already exists within the carbon box model node list,
        then it is not added. If the node is not a Box Class Instance, then it raises a ValueError.
        Parameters
        ----------
        nodes : list | ndarray
            A list of nodes of Type Box to add to the carbon box model.
        Raises
        ------
        ValueError
            If any of the objects in the list are not of type Box.
        """
        for node in nodes:
            if isinstance(node, Box):
                if node in self._nodes.values():
                    print(node.get_name() + " already exists in the Carbon Box. It was ignored and not "
                                            "added to existing nodes.")
                else:
                    self._nodes[self._n_nodes] = node
                    self._reverse_nodes[node] = self._n_nodes
                    self._n_nodes += 1
            else:
                raise ValueError("One/many of the input nodes are not of Box Class.")

    def add_edges(self, flow_objs):
        """ Adds the flow objects specified in the list to the carbon box model. If any of the objects in the list are
            not an instance of Flow Class then it throws a ValueError.
        Parameters
        ----------
        flow_objs : list | ndarray
            A list of Flow objects to add to the Carbon Box Model.
        Raises
        ------
        ValueError
            If any of the objects in the list are not of type Flow.
        """
        for flow_obj in flow_objs:
            if not isinstance(flow_obj, Flow):
                raise ValueError("One/many of the input edge are not of Flow Class.")
            self._edges.append(flow_obj)

    def get_edges(self):
        """ Getter method for the name of edges.
        Returns
        -------
        list
            A list of the edges (given in their string representations i.e. str(source) --> str(destination):flux_value).
        """
        return list(map(str, self._edges))

    def get_edges_objects(self):
        """ Getter method for the edge objects themselves.
        Returns
        -------
        list
            A list of Flow Objects that have been added so far to the Class Object.
        """
        return self._edges

    def get_nodes(self):
        """ Getter method for the name of the nodes in the order they were added to the Carbon Box Model Object.
        Returns
        -------
        list
            A list of node names in the order they were inserted.
        """
        return [self._nodes[j].get_name() for j in range(self._n_nodes)]

    def get_nodes_objects(self):
        """ Getter method for the node objects in the order they were inserted.
        Returns
        -------
        list
            A list of the node objects in the order they were inserted.
        """
        return [self._nodes[j] for j in range(self._n_nodes)]

    def get_fluxes(self):
        """ Getter method for the compiled fluxes in the units specified in the init_method. If the compile method has
            not been run, then it will return None.
        Returns
        -------
        DeviceArray
            2D jax numpy array which contains the fluxes where index (i,j) indicates the flux from node[i] to node[j].
        """
        return self._fluxes

    def get_converted_fluxes(self):
        """ Getter method for the fluxes when converted to 'Gt/yr' (this is the unit that the rest of the methods work
            in internally). This returns None if the compile method has not been run.
        Returns
        -------
        DeviceArray
            2D jax numpy array which contains the unit-corrected fluxes where index (i,j) indicates the 'Gt/yr' flux
            from node[i] to node[j].
        """
        return self._corrected_fluxes

    def get_reservoir_contents(self):
        """ Getter method for the 12C reservoir content of the nodes (Boxes).. Returns None if the compile method has
        not been called.
        Returns
        -------
        DeviceArray
            jax numpy array containing the reservoir content of the nodes (returned in the order the nodes were added).
        """
        return self._reservoir_content

    def get_production_coefficients(self):
        """ Getter method for the normalised production coefficients of the nodes (Boxes). Returns None if compile method
        has not been called.
        Returns
        -------
        DeviceArray
            jax array containing the normalised production coefficients of the nodes (returned in the order the
            nodes were added).
        """
        return self._production_coefficients


    def get_matrix(self):
        """ Getter method for the ODE coefficient matrix to solve.
        Returns
        -------
        DeviceArray
            jax array of the ODE coefficient matrix to solve.
        """
        return self._matrix

    @partial(jit, static_argnums=0)
    def _convert_production_rate(self, production_rate):
        if self._production_rate_units == 'atoms/cm^2/s':
            new_rate = production_rate * 14.003242 / 6.022 * 5.11 * 31536. / 1.e5
        elif self._production_rate_units == 'kg/yr':
            new_rate = production_rate
        else:
            raise ValueError('Production Rate units must be either atoms/cm^2/s or kg/yr!')
        return new_rate

    def _convert_flux_rate(self, fluxes):
        if self._flow_rate_units == 'Gt/yr':
            corrected_fluxes = self._fluxes
        elif self._flow_rate_units == '1/yr':
            corrected_fluxes = fluxes * jnp.transpose(self._reservoir_content) * 14.003242 / 12
            corrected_fluxes = corrected_fluxes + jnp.transpose(corrected_fluxes)
        else:
            raise ValueError('Flow rate units must be either Gt/yr or 1/yr!')
        return corrected_fluxes

    def compile(self):
        """ Method which compiles crucial parts of the model. If the model has not been compiled before then it
        compiles the following quantities:
        - 12C reservoir content of the nodes (in the order the nodes were added)
        - the fluxes (where fluxes at index (i,j) represents the flux from node[i] to node[j])
        - the decay matrix (matrix with decay constant along the diagonal)
        - production coefficients of the nodes.
        - the corrected fluxes (fluxes in unit 'Gt/yr')
        - the matrix of the coefficients for the ODEINT to solve.
        It also detects if the incoming and outgoing fluxes at every node is balanced and if not then throws ValueError
        along with which node is unbalanced.
        Raises
        ------
        ValueError
            If the incoming flux and outgoing flux at every Box is not balanced.
        """

        if self._fluxes is None:

            self._reservoir_content = jnp.array(
                [[self._nodes[j].get_reservoir_content() for j in range(self._n_nodes)]])
            self._fluxes = jnp.zeros((self._n_nodes, self._n_nodes))
            for flow in self._edges:
                self._fluxes = self._fluxes.at[self._reverse_nodes[flow.get_source()], 
                    self._reverse_nodes[flow.get_destination()]].set(flow.get_flux())

            self._decay_matrix = jnp.diag(jnp.array([self._decay_constant] * self._n_nodes))
            self._production_coefficients = jnp.array([self._nodes[j].get_production() for j in range(self._n_nodes)])
            self._production_coefficients /= jnp.sum(self._production_coefficients)
            self._corrected_fluxes = self._convert_flux_rate(self._fluxes)
            c_14_fluxes = self._corrected_fluxes / jnp.transpose(self._reservoir_content)
            new_c_14_fluxes = jnp.diag(jnp.sum(c_14_fluxes, axis=1))
            self._matrix = jnp.transpose(c_14_fluxes) - new_c_14_fluxes - self._decay_matrix
            self._non_hemisphere_model = self._nodes[0].get_hemisphere() == "None"

            for i in range(self._n_nodes):
                if jnp.abs(jnp.sum(self._corrected_fluxes[:, i]) - jnp.sum(self._corrected_fluxes[i, :])) > 0.001:
                    raise ValueError('the outgoing and incoming fluxes are not balanced for ' + str(self._nodes[i]))
                if self._non_hemisphere_model:
                    if not self._nodes[i].get_hemisphere() == "None":
                        raise ValueError(str(self._nodes[i]) + ' has been given a hemisphere value when there are '
                                                               'others which are None!')
                else:
                    if not self._nodes[i].get_hemisphere() in ['north', 'south']:
                        raise ValueError('There are some nodes where hemisphere is given and others where it is not. '
                                         'This is not allowed either all must be None of all of them must be either '
                                         '"north" or "south"')

    def _equilibrate_brehm(self, production_rate):
        production_rate = self._convert_production_rate(production_rate)
        solution = jnp.linalg.solve(self._matrix, -1 * self._production_coefficients * production_rate)
        return solution

    def _equilibrate_guttler(self, target_C_14):
        troposphere_index = None
        for index, node in self._nodes.items():
            if node.get_name() == 'Troposphere':
                troposphere_index = index
                break
        if troposphere_index is None:
            raise ValueError('there is currently no Troposphere node to equilibrate!')

        @jit
        def objective_function(production_rate):
            """ Function which calculates the difference between the current tropospheric C14 content and the target
            tropospheric C14 content.
            Parameters
            ----------
            production_rate : float
            Returns
            -------
            float
                The difference between the target and the equilibrium production.
            """
            equilibrium = self._equilibrate_brehm(production_rate)
            return (equilibrium[troposphere_index] - target_C_14) ** 2

        grad_obj = jax.jit(jax.grad(objective_function))
        final_production_rate = scipy.optimize.minimize(objective_function, jnp.array([6.]), method='BFGS',
                                                        jac=grad_obj)
        return final_production_rate.x[0]

    def equilibrate(self, target_C_14=None, production_rate=None):
        """ External equilibrate method which determines the appropriate result to return given a parameter. If
        neither parameter is given then it throws a ValueError. If both are specified, then it ignores production_rate.
        Parameters
        ----------
        target_C_14 : float
            target C14 with which to equilibrate with. Defaults to None.
            
        production_rate : float
            production rate with which to equilibrate to. Defaults to None.
        Returns
        -------
        float | ndarray
            if target C14 is specified, it returns the initial production rate. If the production rate is specified,
            then it returns a ndarray of the C14 reservoir content.
        Raises
        ------
        ValueError
            If the both arguments are specified as None.
        """

        if target_C_14 is not None:
            return self._equilibrate_guttler(target_C_14)

        elif production_rate is not None:
            return self._equilibrate_brehm(production_rate)

        else:
            raise ValueError("Must give either target C-14 or production rate.")

    @partial(jit, static_argnums=(0, 2, 5, 6, 8))
    def run(self, time, production, y0=None, args=(), target_C_14=None, steady_state_production=None, solution=None,adaptive=True,step_ts=None):
        """ For the given production function, this calculates the C14 content of all the boxes within the carbon box
        model at the specified time values. It does this by solving a linear system of ODEs. This method will not work
        if the compile method has not been executed first.
        Parameters
        ----------
        time : list
            the time values at which to calculate the content of all the boxes.
        production : callable
            the production function which determines the contents of the boxes.
        y0 : list, optional
            the initial contents of all boxes. Defaults to None. Must be a length(n) list of the same size as the number
            of boxes.
        args : tuple, optional
            optional arguments to pass into the production function.
        steady_state_production : int, optional
            the steady state production rate with which to equilibrate with to find steady state solution. If y0 is
            specified, this parameter is ignored.
        target_C_14 : int, optional
            target C14 with which to equilibrate with to find steady state solution. If y0 or steady_state_production is
             specified, this parameter is ignored.
        solution : ndarray, optional
            the equilibrium solution to the ODE system. If this is not specified, then it will be calculated internally.
            Must be of the same length as the number of boxes.
        Returns
        -------
        Union[list, list]
            The value of each box in the carbon box at the specified time_values along with the steady state solution
            for the system.
        Raises
        ------
        ValueError
            If neither the target C-14 nor production rate is specified.
            If the production is not a callable function.
        """

        def derivative(t, y, args):
            ans = jnp.matmul(self._matrix, y)
            production_rate_constant = production(t, *args) - steady_state_production
            production_rate_constant = self._convert_production_rate(production_rate_constant)
            production_term = self._production_coefficients * production_rate_constant
            return ans + production_term

        time_values = jnp.array(time)
        if solution is None:
            if steady_state_production is not None:
                solution = self.equilibrate(production_rate=steady_state_production)

            elif target_C_14 is not None:
                steady_state_production = self.equilibrate(target_C_14=target_C_14)
                solution = self.equilibrate(production_rate=steady_state_production)
            else:
                ValueError("Must give either target C-14 or production rate or steady state values of system.")

        if not callable(production):
            raise ValueError("incorrect object type for production")
            
        if y0 is not None:
            y_initial = jnp.array(y0)
        else:
            y_initial = jnp.array(solution)

        term = diffrax.ODETerm(derivative)
        solver = diffrax.Dopri5()
        saveat = diffrax.SaveAt(ts=time_values)

        # for fixed time sampling
        start = time_values[0]
        end = time_values[-1]
        step = 1 / 48 # 4 per month 
        max_steps = None #48*jnp.size(time_values)
        
        if adaptive:
            stepsize_controller = diffrax.PIDController(rtol=5e-11, atol=5e-11,
                dtmin=step/2.,dtmax=step*4,force_dtmin=True,step_ts=step_ts)

        else:
            stepsize_controller = diffrax.ConstantStepSize()
        dt0 = step

        states = diffrax.diffeqsolve(term, solver, args=args, y0=y_initial-solution, t0 = start, t1 = end,
            dt0 = dt0, stepsize_controller=stepsize_controller,saveat=saveat, 
            max_steps=max_steps)

        return states.ys+solution, solution

    @partial(jit, static_argnums=(0, 2))
    def bin_data(self, data, time_oversample, time_out, growth):
        """ Bins the data given based on the oversample and the growth season according to Schulman's convention.
        Can handle any contiguous growth season, even over the year.
        Parameters
        ----------
        data : ndarray
            the data which to bin.
        time_oversample : int
            number of samples taken per year.
        time_out : ndarray
            the time values at which to bin the data at.
        growth : ndarray
            the growth season with which to bin the data with respect to.
        Returns
        -------
        DeviceArray
            The final binned data accounting for both growth season and Schulman's convention.
        Raises
        ------
        ValueError
            If the data is not one-dimensional or in a single row.
        """
        if data.ndim != 1:
            raise ValueError("Data is not one-dimensional! Data must be contained in one row. ")

        masked = jnp.linspace(0, 1, time_oversample)
        kernel = (masked < jnp.count_nonzero(growth)/12)
        
        @partial(jit)
        def _shifted_index_finder(seasons):
            first1 = jnp.where(seasons == 1, size=1)[0][0]
            first0 = jnp.where(seasons == 0, size=1)[0][0]
            all1s = jnp.where(seasons == 1, size=12)[0]
            after1 = jnp.where(all1s > first0, all1s, 0)
            after1 = after1.at[jnp.nonzero(after1, size=1)].get()[0]
            num = jax.lax.sub(first1, after1)
            val = cond(num == 0, lambda x: first1, lambda x : after1, num)
            act = cond(jnp.all(seasons == 1), lambda x: 0, lambda x: val, seasons)
            return act
    
        shifted_index = _shifted_index_finder(growth)
  
        binned_data = self._rebin1D(time_out, shifted_index, time_oversample, kernel, data)
        return binned_data

    @partial(jit, static_argnums=(0, 3))
    def _rebin1D(self, time_out, shifted_index, oversample, kernel, s):
        binned_data = jnp.zeros((len(time_out),))
        fun = lambda i, val: dynamic_update_slice(val, jnp.array([jnp.sum(jnp.multiply(dynamic_slice(
            s, (i * oversample + shifted_index * oversample // 12,), (oversample,)), kernel)) / (
                                                                      jnp.sum(kernel))]), (i,))
        binned_data = fori_loop(0, len(time_out), fun, binned_data)
        return binned_data


def save_model(carbon_box_model, filename):
    """ Saves the Carbon Box Model in a hd5 format with the specified filename. If the first parameter is not of Type
    CarbonBoxModel, then it throws a ValueError.
    Parameters
    ----------
    carbon_box_model : CarbonBoxModel
        the model to save.
    filename : str
        file name and location where Carbon Box Model needs to be saved. Must have a '.hd5' at end of filename.
    Raises
    ------
    ValueError
        If the first parameter is not of type CarbonBoxModel.
    """
    file = h5py.File(filename, 'a')
    file.clear()  # overwrites if it already exists
    if isinstance(carbon_box_model, CarbonBoxModel):
        metadata = {'fluxes': carbon_box_model.get_converted_fluxes(),
                    'reservoir content': carbon_box_model.get_reservoir_contents(),
                    'production coefficients': carbon_box_model.get_production_coefficients(),
                    'nodes': carbon_box_model.get_nodes(),
                    'hemispheres': [i.get_hemisphere() for i in carbon_box_model.get_nodes_objects()]}
    else:
        raise ValueError("parameter is not a Carbon Box Model!")

    hdfdict.dump(metadata, file)
    file.close()


def load_model(filename, production_rate_units='kg/yr', flow_rate_units='Gt/yr'):
    """ Loads the saved Carbon Box Model from the relevant filename. Units for both production rate and flow rate can be
    specified as parameters. filename must be specified with the .hd5 extension.
    Parameters
    ----------
    filename : str
        the name and location (in one string) of the file where the Carbon Box Model is saved.
    production_rate_units : str, optional
        the production rate of the model to be loaded. Defaults to 'kg/yr'.
    flow_rate_units : str, optional
        the production rate of the model to be loaded. Defaults to 'Gt/yr'.
    Returns
    -------
    CarbonBoxModel
        Carbon Box Model which is generated from the file.
    Raises
    ------
    ValueError
        If the flow rate units are not either 'Gt/yr' or '1/yr'.
    """

    file = h5py.File(filename, 'r')
    metadata = dict(hdfdict.load(file))
    nodes = metadata['nodes']
    hemispheres = metadata['hemispheres']
    carbon_box_model = CarbonBoxModel(production_rate_units=production_rate_units, flow_rate_units=flow_rate_units)
    box_object_list = []
    for j in range(len(nodes)):
        box_object_list.append(Box(nodes[j].decode("utf-8"), float(metadata['reservoir content'][0][j]),
                                   float(metadata['production coefficients'][j]), hemispheres[j].decode("utf-8")))

    carbon_box_model.add_nodes(box_object_list)

    for k in range(metadata['fluxes'].shape[0]):
        for j in range(metadata['fluxes'].shape[1]):
            if metadata['fluxes'][k, j] != 0:
                if flow_rate_units == 'Gt/yr':
                    carbon_box_model.add_edges([Flow(box_object_list[k], box_object_list[j],
                                                     float(metadata['fluxes'][k, j]))])

                elif flow_rate_units == '1/yr':
                    new_flow = float(metadata['fluxes'][k, j]) * 12 / 14.003242 / box_object_list[
                        k].get_reservoir_content()
                    carbon_box_model.add_edges([Flow(box_object_list[k], box_object_list[j], new_flow)])
                else:
                    raise ValueError('flow_rate_units are not valid.')

    file.close()
    return carbon_box_model


def load_presaved_model(model, production_rate_units='kg/yr', flow_rate_units='Gt/yr'):
    """ Loads a pre-saved, commonly used model based on the research papers linked below. The model must be one of the
    following: Miyake17, Brehm21, Guttler15, Buntgen18. Loads the model based on the units for production rate
    and flow rate specified.
    Parameters
    ----------
    model : str
        the name of the model to load. Must be one in [Miyake17, Brehm21, Guttler15, Buntgen18].
    production_rate_units : str, optional
        the production rate of the model to be loaded. Defaults to 'kg/yr'.
    flow_rate_units : str, optional
        the production rate of the model to be loaded. Defaults to 'Gt/yr'.
    Returns
    -------
    CarbonBoxModel
        Carbon Box Model which is generated from the pre-saved file.
    Raises
    ------
    ValueError
        If the specified model parameter is not in the required list.
    """
    if model in ['Guttler15', 'Brehm21', 'Miyake17', 'Buntgen18']:
        file = 'data/' + model + '.hd5'
        carbonmodel = load_model(pkg_resources.resource_stream(__name__, file),
                                 production_rate_units=production_rate_units, flow_rate_units=flow_rate_units)
        return carbonmodel
    else:
        raise ValueError('model parameter must be one of the following: Guttler15, Brehm21, Miyake17, Buntgen18')



from ticktack import Box, Flow, CarbonBoxModel, save_model

# ------------- Guttler15 Model ------------- #
SS = Box("Sedimentary Sink", 378000)
Sw = Box("Surface Water", 900)
Sb = Box("Surface Biota", 3)
Idw = Box("Intermediate & Deep water", 37800)
Slb = Box("Short-lived biota", 110)
Llb = Box("Long-lived Biota", 450)
L = Box("Litter", 300)
S = Box("Soil", 1350)
P = Box("Peat", 500)
Sp = Box("Stratosphere", 88.5, production_coefficient=0.7)
Tp = Box("Troposphere", 501.5, production_coefficient=0.3)

f1 = Flow(SS, Tp, 0.7)
f2 = Flow(Sp, Tp, 45)
f3 = Flow(Tp, Sp, 45)
f4 = Flow(Tp, Sw, 60.5)
f5 = Flow(Sw, Tp, 61)
f6 = Flow(Sw, Sb, 40)
f7 = Flow(Sb, Sw, 36)
f8 = Flow(Sb, Idw, 4)
f9 = Flow(Idw, Sw, 42)
f10 = Flow(Sw, Idw, 38.2)
f11 = Flow(Sw, SS, 0.3)
f12 = Flow(Idw, SS, 0.2)
f13 = Flow(Tp, Slb, 115)
f14 = Flow(Slb, Tp, 60)
f15 = Flow(Slb, Llb, 15)
f16 = Flow(Llb, L, 15)
f17 = Flow(Slb, L, 40)
f18 = Flow(L, S, 3)
f19 = Flow(L, P, 1)
f20 = Flow(L, Sw, 1)
f21 = Flow(L, Tp, 50)
f22 = Flow(S, Tp, 3)
f23 = Flow(P, Tp, 0.8)
f24 = Flow(P, SS, 0.2)
CBM = CarbonBoxModel(production_rate_units='atoms/cm^2/s')
to_add = [Sp, Tp, Sw, Sb, Idw, Slb, Llb, L, S, P, SS]
CBM.add_nodes(to_add)
CBM.add_edges([f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11, f12, f13, f14, f15, f16, f17, f18, f19, f20,
               f21, f22, f23, f24])
CBM.compile()
save_model(CBM, 'data/Guttler15.hd5')


# ------------- Brehm21 Model ------------- #
SSS = Box("Sedimentary Sink", 189000, hemisphere='south')
SwS = Box("Surface Water", 540,  hemisphere='south')
SbS = Box("Surface Biota", 1.8,  hemisphere='south')
IdwS = Box("Intermediate and Deep water", 22680, hemisphere='south')
SlbS = Box("Short-lived biota", 33, hemisphere='south')
LlbS = Box("Long-lived Biota", 135, hemisphere='south')
LS = Box("Litter", 90, hemisphere='south')
SoS = Box("Soil", 405, hemisphere='south')
PS = Box("Peat", 150, hemisphere='south')
SpS = Box("Stratosphere", 44.5, production_coefficient=0.7*0.5, hemisphere='south')
TpS = Box("Troposphere", 250.5, production_coefficient=0.3*0.5, hemisphere='south')
SSN = Box("Sedimentary Sink", 189000, hemisphere='north')
SwN = Box("Surface Water", 360,hemisphere='north')
SbN = Box("Surface Biota", 1.2,hemisphere='north')
IdwN = Box("Intermediate & Deep water", 15120,hemisphere='north')
SlbN = Box("Short-lived biota", 77,hemisphere='north')
LlbN = Box("Long-lived Biota", 315,hemisphere='north')
LN = Box("Litter", 210,hemisphere='north')
SoN = Box("Soil", 945,hemisphere='north')
PN = Box("Peat", 350,hemisphere='north')
SpN = Box("Stratosphere", 44.5, production_coefficient=0.7*0.5,hemisphere='north')
TpN = Box("Troposphere", 250.5, production_coefficient=0.3*0.5,hemisphere='north')

f1 = Flow(SpS, TpS, 22.5)
f2 = Flow(TpS, SpS, 22.5)
f3 = Flow(TpS, SlbS, 34.5)
f4 = Flow(SlbS, TpS, 18)
f5 = Flow(TpS, SwS, 36.6)
f6 = Flow(SwS, TpS, 36.6)
f7 = Flow(SlbS, LlbS, 4.5)
f8 = Flow(LlbS, LS, 4.5)
f9 = Flow(SlbS, LS, 12)
f10 = Flow(LS, TpS, 15)
f11 = Flow(LS, SwS, 0.3)
f12 = Flow(LS, PS, 0.3)
f13 = Flow(LS, SoS, 0.9)
f14 = Flow(PS, TpS, 0.24)
f15 = Flow(PS, SSS, 0.06)
f16 = Flow(SoS, TpS, 0.9)
f17 = Flow(SwS, SSS, 0.18)
f18 = Flow(IdwS, SSS, 0.12)
f19 = Flow(SwS, SbS, 22.6)
f20 = Flow(SwS, IdwS, 37.92)
f21 = Flow(SwS, SwN, 13)
f22 = Flow(SbS, SwS, 20.2)
f23 = Flow(SbS, IdwS, 2.4)
f24 = Flow(IdwS, SwS, 43.2)
f25 = Flow(IdwS, IdwN, 15.12)
f26 = Flow(SSS, TpS, 0.36)
f27 = Flow(SpS, SpN, 44.1)
f28 = Flow(TpS, TpN, 71)
f29 = Flow(SpN, SpS, 44.1)
f30 = Flow(TpN, TpS, 71)
f31 = Flow(SpN, TpN, 22.5)
f32 = Flow(TpN, SpN, 22.5)
f33 = Flow(TpN, SwN, 24.4)
f34 = Flow(TpN, SlbN, 80.5)
f35 = Flow(SwN, TpN, 24.89)
f36 = Flow(SwN, SwS, 10)
f37 = Flow(SwN, SbN, 16)
f38 = Flow(SwN, IdwN, 28.29)
f39 = Flow(SlbN, TpN, 42)
f40 = Flow(SlbN, LlbN, 10.5)
f41 = Flow(SlbN, LN, 28)
f42 = Flow(LlbN, LN, 10.5)
f43 = Flow(SbN, SwN, 14.4)
f44 = Flow(SbN, IdwN, 1.6)
f45 = Flow(SwN, SSN, 0.12)
f46 = Flow(IdwN, SwN, 26.8)
f47 = Flow(IdwN, SSN, 0.09)
f48 = Flow(IdwN, IdwS, 18.12)
f49 = Flow(LN, SwN, 0.7)
f50 = Flow(LN, SoN, 2.1)
f51 = Flow(LN, TpN, 35)
f52 = Flow(LN, PN, 0.7)
f53 = Flow(SoN, TpN, 2.1)
f54 = Flow(PN, TpN, 0.56)
f55 = Flow(PN, SSN, 0.14)
f56 = Flow(SSN, TpN, 0.35)
CBM = CarbonBoxModel()
to_add = [SpS, TpS, SwS, SbS, IdwS, SlbS, LlbS, LS, SoS,
          PS, SSS, SpN, TpN, SwN, SbN, IdwN, SlbN, LlbN, LN, SoN, PN, SSN]
CBM.add_nodes(to_add)
CBM.add_edges([f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11, f12, f13, f14, f15, f16, f17, f18, f19, f20,
               f21, f22, f23, f24, f25, f26, f27, f28, f29, f30, f31, f32, f33, f34, f35, f36, f37, f38, f39, f40,
               f41, f42, f43, f44, f45, f46, f47, f48, f49, f50, f51, f52, f53, f54, f55, f56])
CBM.compile()
save_model(CBM, 'data/Brehm21.hd5')


# ------------- Buntgen18 Model ------------- #
SSS = Box("Sedimentary Sink", 189000, hemisphere='south')
SwS = Box("Surface Water", 540, hemisphere='south')
SbS = Box("Surface Biota", 1.8, hemisphere='south')
IdwS = Box("Intermediate and Deep water", 22680, hemisphere='south')
SlbS = Box("Short-lived biota", 33, hemisphere='south')
LlbS = Box("Long-lived Biota", 135, hemisphere='south')
LS = Box("Litter", 90, hemisphere='south')
SoS = Box("Soil", 405, hemisphere='south')
PS = Box("Peat", 150, hemisphere='south')
SpS = Box("Stratosphere", 44.5, production_coefficient=0.7*0.5, hemisphere='south')
TpS = Box("Troposphere", 250.5, production_coefficient=0.3*0.5, hemisphere='south')
SSN = Box("Sedimentary Sink", 189000, hemisphere='north')
SwN = Box("Surface Water", 360, hemisphere='north')
SbN = Box("Surface Biota", 1.2, hemisphere='north')
IdwN = Box("Intermediate & Deep water", 15120, hemisphere='north')
SlbN = Box("Short-lived biota", 77, hemisphere='north')
LlbN = Box("Long-lived Biota", 315, hemisphere='north')
LN = Box("Litter", 210, hemisphere='north')
SoN = Box("Soil", 945, hemisphere='north')
PN = Box("Peat", 350, hemisphere='north')
SpN = Box("Stratosphere", 44.5, production_coefficient=0.7*0.5, hemisphere='north')
TpN = Box("Troposphere", 250.5, production_coefficient=0.3*0.5, hemisphere='north')

f1 = Flow(SpS, TpS, 23.8)
f2 = Flow(TpS, SpS, 23.8)
f3 = Flow(TpS, SlbS, 35.36)
f4 = Flow(SlbS, TpS, 18.86)
f5 = Flow(TpS, SwS, 38.25)
f6 = Flow(SwS, TpS, 38.25)
f7 = Flow(SlbS, LlbS, 4.5)
f8 = Flow(LlbS, LS, 4.5)
f9 = Flow(SlbS, LS, 12)
f10 = Flow(LS, TpS, 15)
f11 = Flow(LS, SwS, 0.3)
f12 = Flow(LS, PS, 0.3)
f13 = Flow(LS, SoS, 0.9)
f14 = Flow(PS, TpS, 0.24)
f15 = Flow(PS, SSS, 0.06)
f16 = Flow(SoS, TpS, 0.9)
f17 = Flow(SwS, SSS, 0.18)
f18 = Flow(IdwS, SSS, 0.12)
f19 = Flow(SwS, SbS, 22.6)
f20 = Flow(SwS, IdwS, 8.52)
f21 = Flow(SwS, SwN, 28.8)
f22 = Flow(SbS, SwS, 20.2)
f23 = Flow(SbS, IdwS, 2.4)
f24 = Flow(IdwS, SwS, 39.6)
f25 = Flow(IdwS, IdwN, 0.72)
f26 = Flow(SSS, TpS, 0.36)
f27 = Flow(SpS, SpN, 22.25)
f28 = Flow(TpS, TpN, 71.6)
f29 = Flow(SpN, SpS, 22.25)
f30 = Flow(TpN, TpS, 71.6)
f31 = Flow(SpN, TpN, 23.8)
f32 = Flow(TpN, SpN, 23.8)
f33 = Flow(TpN, SwN, 29.16)
f34 = Flow(TpN, SlbN, 80.51)
f35 = Flow(SwN, TpN, 29.65)
f36 = Flow(SwN, SwS, 0)
f37 = Flow(SwN, SbN, 16)
f38 = Flow(SwN, IdwN, 49.69)
f39 = Flow(SlbN, TpN, 42.01)
f40 = Flow(SlbN, LlbN, 10.5)
f41 = Flow(SlbN, LN, 28)
f42 = Flow(LlbN, LN, 10.5)
f43 = Flow(SbN, SwN, 14.4)
f44 = Flow(SbN, IdwN, 1.6)
f45 = Flow(SwN, SSN, 0.12)
f46 = Flow(IdwN, SwN, 22.4)
f47 = Flow(IdwN, SSN, 0.09)
f48 = Flow(IdwN, IdwS, 29.52)
f49 = Flow(LN, SwN, 0.7)
f50 = Flow(LN, SoN, 2.1)
f51 = Flow(LN, TpN, 35)
f52 = Flow(LN, PN, 0.7)
f53 = Flow(SoN, TpN, 2.1)
f54 = Flow(PN, TpN, 0.56)
f55 = Flow(PN, SSN, 0.14)
f56 = Flow(SSN, TpN, 0.35)
CBM = CarbonBoxModel()
to_add = [SpS, TpS, SwS, SbS, IdwS, SlbS, LlbS, LS, SoS,
          PS, SSS, SpN, TpN, SwN, SbN, IdwN, SlbN, LlbN, LN, SoN, PN, SSN]
CBM.add_nodes(to_add)
CBM.add_edges([f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11, f12, f13, f14, f15, f16, f17, f18, f19, f20,
               f21, f22, f23, f24, f25, f26, f27, f28, f29, f30, f31, f32, f33, f34, f35, f36, f37, f38, f39, f40,
               f41, f42, f43, f44, f45, f46, f47, f48, f49, f50, f51, f52, f53, f54, f55, f56])
CBM.compile()
save_model(CBM, 'data/Buntgen18.hd5')


# ------------- Miyake17 Model ------------- #
strat = Box('Stratosphere', 112.5, 0.7)
trop = Box('Troposphere', 637.5, 0.3)
MS = Box("Marine surface", 900)
Bio = Box("Biosphere", 1600)
f1 = Flow(strat, trop, 1 / 2)
f2 = Flow(trop, MS, 1 / 11)
f3 = Flow(trop, Bio, 1 / 23)
CBM = CarbonBoxModel(flow_rate_units='1/yr')
CBM.add_nodes([strat, trop, MS, Bio])
CBM.add_edges([f1, f2, f3])
CBM.compile()
save_model(CBM, 'data/Miyake17.hd5')


from setuptools import setup, find_packages
import os
import codecs
import re
long_description = 'A Python package implementing carbon box models for tree ring radiocarbon data analysis.'

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# DEPENDENCIES
# 1. What are the required dependencies?
with open('requirements.txt') as f:
    install_requires = f.read().splitlines()
# 2. What dependencies required to run the unit tests? (i.e. `pytest --remote-data`)
tests_require = ['pytest', 'pytest-cov', 'pytest-remotedata']

setup(name='ticktack',
      version=find_version("src", "ticktack", "__init__.py"),
      description='Python package to build a carbon box model.',
      long_description=long_description,
      author='Utkarsh Sharma',
      author_email='u.sharma@uqconnect.edu.au',
      url='https://github.com/SharmaLlama/ticktack',
      package_dir={'': 'src'},
      packages=find_packages(where='src'),
      package_data={'': ['data/*.hd5', 'data/datasets/660BCE-Ew/*.csv', 'data/datasets/660BCE-Lw/*.csv',
                         'data/datasets/775AD-early-N/*.csv', 'data/datasets/775AD-early-S/*.csv',
                         'data/datasets/775AD-late-N/*.csv',
                         'data/datasets/993AD-N/*.csv', 'data/datasets/993AD-S/*.csv',
                         'data/datasets/5259BCE/*.csv', 'data/datasets/5410BCE/*.csv',
                         'data/datasets/7176BCE/*.csv'
                         ]},
      include_package_data=True,
      install_requires=install_requires,
      tests_require=tests_require,
      extras_require={'tests': tests_require},
      license='MIT',
      classifiers=[
          "Topic :: Scientific/Engineering",
          "Intended Audience :: Science/Research",
          "Intended Audience :: Developers",
          "Development Status :: 4 - Beta",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Programming Language :: Python"
      ]
      )



import jax.numpy as jnp
from jax import grad, jit, random, jacrev, vmap
from jax.lax import cond, sub

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import matplotlib as mpl
from matplotlib.ticker import MaxNLocator
import scipy

import ticktack

import diffrax 
from jax_cosmo.scipy.interpolate import InterpolatedUnivariateSpline
from tinygp import kernels, GaussianProcess

from functools import partial

from astropy.table import Table
from tqdm import tqdm
import emcee

from chainconsumer import ChainConsumer

import seaborn as sns
# from jaxns.nested_sampler import NestedSampler
# from jaxns.prior_transforms import PriorChain, UniformPrior

import os



class CarbonFitter:
    """
    Parent class of SingleFitter and MultiFitter. Does Monte Carlo sampling, plotting and more.
    """

    def MarkovChainSampler(self, params, likelihood, burnin=500, production=1000, k=2, args=()):
        """
        Runs an affine-invariant MCMC sampler on an array of initial parameters, subject to some likelihood function.
        Parameters
        ----------
        params : ndarray
            Initial parameters for MC sampler
        likelihood : callable
            Log-likelihood function for params
        burnin : int, optional
            Number of steps to run in burn-in period. 500 by default.
        production : int, optional
            Number of steps to run in production period. 1000 by default.
        k: int, optional
            Number of walkers per parameter. 2 by default.
        Returns
        -------
        ndarray
            A chain of MCMC walk
        """
        initial = params
        ndim, nwalkers = len(initial), k * len(initial)
        sampler = emcee.EnsembleSampler(nwalkers, ndim, likelihood, args=args)

        print("Running burn-in...")
        p0 = initial + 1e-5 * np.random.rand(nwalkers, ndim)
        p0, lp, _ = sampler.run_mcmc(p0, burnin, progress=True);

        print("Running production...")
        sampler.reset()
        sampler.run_mcmc(p0, production, progress=True);
        return sampler.get_chain(flat=True)

    # def NestedSampler(self, params, likelihood, low_bound=None, high_bound=None, sampler_name='multi_ellipsoid'):
    #     """
    #     Runs Nested Sampling sampler on the parameter space of some model, subject to some likelihood function.
    #     Parameters
    #     ----------
    #     params : ndarray
    #         Initial parameters for NS sampler
    #     likelihood : callable
    #         Log-likelihood function for params
    #     low_bound : ndarray, optional
    #         Lower bound of params
    #     high_bound : ndarray, optional
    #         Upper bound of params
    #     sampler_name : str, optional
    #         The sampling method for NS sampler. Must be one in: 'multi_ellipsoid', 'slice'. 'multi_ellipsoid' by default.
    #     Returns
    #     -------
    #     ndarray
    #         A chain of NS samples
    #     """
    #
    #     @jit
    #     def likelihood_function(params, **kwargs):
    #         return likelihood(params)
    #
    #     ndim = params.size
    #     if low_bound is not None and high_bound is not None:
    #         low_bound = jnp.array(low_bound)
    #         high_bound = jnp.array(high_bound)
    #     else:
    #         low_bound = jnp.array(ndim * None)
    #         high_bound = jnp.array(ndim * None)
    #     prior_chain = PriorChain().push(UniformPrior('params', low=low_bound, high=high_bound))
    #     ns = NestedSampler(likelihood_function, prior_chain, num_live_points=100 * prior_chain.U_ndims,
    #                        sampler_name=sampler_name)
    #     results = jit(ns)(key=random.PRNGKey(0))
    #     # summary(results)
    #     return results

    def chain_summary(self, chain, walkers, figsize=(10, 10), labels=None, plot_dist=False, test_convergence=False,
                      label_font_size=8, tick_font_size=8, mle=False,usetex=False):
        """
        Runs convergence tests and plots posteriors from a MCMC chain.
        Parameters
        ----------
        chain : ndarray
            A MCMC chain
        walkers : int
            The total number of walkers of the chain
        figsize : tuple, optional
            Output figure size
        labels : list[str], optional
            A list of parameter names
        plot_dist : bool, optional
            If True, plot the marginal distributions of parameters. Else, plot both the marginal distribution
             and the posterior surface
        """
        if labels:
            c = ChainConsumer().add_chain(chain, walkers=walkers, parameters=labels)
        else:
            c = ChainConsumer().add_chain(chain, walkers=walkers)

        if test_convergence:
            gelman_rubin_converged = c.diagnostic.gelman_rubin()
            geweke_converged = c.diagnostic.geweke()
            if gelman_rubin_converged and geweke_converged:
                self.convergence = True
            else:
                self.convergence = False
            print("Convergence: %s" % self.convergence)

        if plot_dist:
            fig = c.plotter.plot_distributions(figsize=figsize)
        else:
            c.configure(spacing=0.0, usetex=usetex, label_font_size=label_font_size, tick_font_size=tick_font_size,
                        diagonal_tick_labels=False)
            fig = c.plotter.plot(figsize=figsize)
            
        if mle:
            MLE = []
            for lst in c.analysis.get_summary().values():
                MLE.append(lst[1])
            return fig, MLE
        else:
            return fig

    def correlation_plot(self, array, figsize=10, square_size=100):
        """
        Makes an accessible heatmap for visualizing correlation/covariance matrix.
        Parameters
        ----------
        array : ndarray
            n x n matrix
        figsize : int, optional
            Output figure size. 10 by default
        square_size: int, optional
            Size of squares on the heatmap. 100 by default
        Returns
        -------
        figure
            heatmap
        """
        assert array.shape[0] == array.shape[1], "array must be a square (n x n) matrix"
        n = array.shape[0]
        arr = array.reshape(-1)
        size = np.abs(arr)
        x = np.repeat(np.arange(n), n)
        y = (n - 1) - np.tile(np.arange(n), n)

        plot_grid = plt.GridSpec(1, 10, hspace=0.2, wspace=0.1)
        ax0 = plt.subplots(figsize=(figsize, figsize))
        ax = plt.subplot(plot_grid[:, :-1])

        n_colors = 256
        palette = sns.diverging_palette(10, 220, n=n_colors)

        if np.min(array) < 0:
            bounds = np.abs(np.max((np.min(array), np.max(array))))
            color_min, color_max = (-bounds, bounds)
        else:
            color_min, color_max = (np.min(array), np.max(array))

        def value_to_color(val):
            val_position = (val - color_min) / (color_max - color_min)
            ind = int(val_position * (n_colors - 1))
            return palette[ind]

        ax.scatter(
            x=x, y=y, s=size * square_size, c=[value_to_color(i) for i in arr], marker='s')
        ax.set_xticks(np.unique(x))
        ax.set_yticks(np.unique(x))

        ax.set_xticklabels(np.unique(x))
        ax.set_yticklabels(reversed(np.unique(x)))

        ax.grid(False, 'major')
        ax.grid(True, 'minor')
        ax.set_xticks([t + 0.5 for t in ax.get_xticks()], minor=True)
        ax.set_yticks([t + 0.5 for t in ax.get_yticks()], minor=True)
        ax.set_xlim([-0.5, (n - 1) + 0.5])
        ax.set_ylim([-0.5, (n - 1) + 0.5])

        ax = plt.subplot(plot_grid[:, -1])
        col_x = [0] * len(palette)
        bar_y = np.linspace(color_min, color_max, n_colors)
        bar_height = bar_y[1] - bar_y[0]
        ax.barh(
            y=bar_y,
            width=[5] * len(palette),
            left=col_x,
            height=bar_height,
            color=palette,
            linewidth=0
        )
        ax.set_xlim(1, 2)
        ax.set_ylim(color_min - 0.001, color_max + 0.001)
        ax.grid(False)
        ax.set_facecolor('white')
        ax.set_xticks([])
        ax.set_yticks([color_min, (color_min + color_max) / 2, color_max])
        ax.yaxis.tick_right()

    def plot_multiple_chains(self, chains, walker, figsize=(10, 10), title=None, params_labels=None, labels=None,
                             colors=None, alpha=0.5, linewidths=None, plot_dists=False, label_font_size=12,
                             tick_font_size=8, max_ticks=10, legend=True, usetex=False, truth=None):
        """
       Overplots posterior surfaces from multiple chains.
        Parameters
        ----------
        chains : list
            List of MCMC chains
        walker : int
            Number of walkers for each chain in 'chains'
        figsize : tuple, optional
            Output figure size
        params_labels : list[str], optional
            List of parameter names
        labels : list[str], optional
            List of labels for different chains
        colors : list[str], optional
            List of colors
        alpha : float, optional
            Parameter for blending, between 0-1.
        linewidths : float, optional
            Line width, in points
        plot_dists : bool, optional
            If True, only plot the marginal distributions of parameters
        label_font_size : int, optional
            Label font size
        tick_font_size : int, optional
            Tick font size
        max_ticks : int, optional
            Maximum number of ticks allowed
        legend : bool, optional
            If True, adds a legend
        Returns
        -------
        figure
            plot of posterior surfaces or marginal distributions
        """
        c = ChainConsumer()
        if labels:
            assert len(labels) == len(chains), "labels must have the same length as chains"
            for i in range(len(chains)):
                c.add_chain(chains[i], walkers=walker, parameters=params_labels, name=labels[i])
        else:
            for i in range(len(chains)):
                c.add_chain(chains[i], walkers=walker, parameters=params_labels)
        c.configure(colors=colors, shade_alpha=alpha, linewidths=linewidths, usetex=usetex,
                    label_font_size=label_font_size, tick_font_size=tick_font_size, diagonal_tick_labels=False,
                    max_ticks=max_ticks)
        # legend_kwargs={"fontsize":14}

        if plot_dists:
            fig = c.plotter.plot_distributions(figsize=figsize, legend=legend)
        else:
            fig = c.plotter.plot(figsize=figsize, legend=legend, truth=truth)
        plt.suptitle(title)
        plt.tight_layout()
        return fig

    def plot_samples(self, average_path=None, chains_path=None, cbm_models=None, hemisphere="north",
                     production_model=None,
                     directory_path=None, size=100, size2=30, alpha=0.05, alpha2=0.2, savefig_path=None):
        colors = mpl.rcParams['axes.prop_cycle'].by_key()['color']
        for i, model in enumerate(cbm_models):
            cbm = ticktack.load_presaved_model(model, production_rate_units='atoms/cm^2/s')
            sf = SingleFitter(cbm, cbm_model=model, hemisphere=hemisphere)
            sf.load_data(average_path)
            chain = np.load(chains_path[i])
            sf.compile_production_model(model=production_model)

            idx = np.random.randint(len(chain), size=size)
            for param in chain[idx]:
                ax1.plot(sf.time_data_fine, sf.dc14_fine(params=param), alpha=alpha, color=colors[i])

            ax1.set_ylabel(r"$\Delta^{14}$C (‰)")
            fig.subplots_adjust(hspace=0.05)

            for param in chain[idx][:size2]:
                ax2.plot(sf.time_data_fine, sf.production(sf.time_data_fine, *param), alpha=alpha2, color=colors[i])

        ax1.errorbar(sf.time_data, sf.d14c_data, yerr=sf.d14c_data_error, fmt="ok", capsize=3,
                     markersize=6.5, elinewidth=3, label=r"average $\Delta^{14}$C")

        if directory_path:
            file_names = [f for f in os.listdir(directory_path) if os.path.isfile(os.path.join(directory_path, f))]
            for file in file_names:
                sf.load_data(directory_path + '/' + file)
                ax1.errorbar(sf.time_data, sf.d14c_data, fmt="o", color="gray", yerr=sf.d14c_data_error, capsize=3,
                             alpha=0.2)
                ax1.plot(sf.time_data, sf.d14c_data, "o", color="gray", alpha=0.2)

        custom_lines = [Line2D([0], [0], color=colors[i], lw=1.5, label=models[i]) for i in range(len(cbm_models))]
        custom_lines.append(Line2D([0], [0], color="k", marker="o", lw=1.5, label=r"average $\Delta^{14}$C"))
        ax1.legend(handles=custom_lines)
        ax2.set_ylim(1, 10)
        ax2.set_xlabel("Calendar Year (CE)")
        ax2.set_xlim(sf.start - 0.2, sf.end + 0.2)
        ax2.set_ylabel("Production rate (atoms/cm$^2$/s)")
        ax2.legend(loc="upper left")
        plt.suptitle(event)
        plt.tight_layout()
        if savefig_path:
            plt.savefig(savefig_path)


class SingleFitter(CarbonFitter):
    """
    A class for parametric and non-parametric inference of d14c data using a Carbon Box Model (CBM).
    Does parameter fitting, Monte Carlo sampling, plotting and more.
    """

    def __init__(self, cbm, cbm_model, production_rate_units='atoms/cm^2/s', target_C_14=707., box='Troposphere',
                 hemisphere='north',adaptive=True):
        """
        Initializes a SingleFitter Object.
        Parameters
        ----------
        cbm : CarbonBoxModel Object
            Carbon Box Model
        cbm_model : str
            CBM name. Must be one in: "Guttler15", "Miyake17", "Buntgen18", "Brehm21"
        production_rate_units : str, optional
            CBM production rate units. 'atoms/cm^2/s' by default
        target_C_14 : float, optional
            Target 14C content for equilibration. 707 by default
        box : str, optional
            Box for calculating d14c. 'Troposphere' by default
        hemisphere : str, optional
            CBM hemisphere. Must be one in: "north", "south". "north" by default
        """
        if isinstance(cbm, str):
            try:
                if cbm in ['Guttler15', 'Brehm21', 'Miyake17', 'Buntgen18']:
                    cbm = ticktack.load_presaved_model(cbm, production_rate_units=production_rate_units)
                else:
                    cbm = ticktack.load_model(cbm, production_rate_units=production_rate_units)
            except:
                raise ValueError('Must be a valid CBM model')
        self.cbm = cbm
        self.cbm.compile()
        self.box = box
        if hemisphere in ['south', 'north']:
            self.hemisphere = hemisphere
        else:
            raise ValueError("'hemisphere' be one of the following: south, north")
        self.cbm_model = cbm_model

        if cbm_model in ['Brehm21', 'Buntgen18']:
            self.steady_state_production = 1.76
            self.steady_state_y0 = self.cbm.equilibrate(production_rate=self.steady_state_production)
            for i, node in enumerate(cbm.get_nodes_objects()):
                if (node.get_hemisphere() == self.hemisphere) & (node.get_name() == self.box):
                    self.box_idx = i
        elif cbm_model == "Miyake17":
            # self.steady_state_production = 1.8
            self.steady_state_production = 1.76
            self.steady_state_y0 = self.cbm.equilibrate(production_rate=self.steady_state_production)
            self.box_idx = 1
        elif cbm_model == "Guttler15":
            # self.steady_state_production = self.cbm.equilibrate(target_C_14=target_C_14)
            self.steady_state_production = 1.76
            self.steady_state_y0 = self.cbm.equilibrate(production_rate=self.steady_state_production)
            self.box_idx = 1

        self.steady_state_box = self.steady_state_y0[self.box_idx]
        self.adaptive = adaptive

    def load_data(self, file_name, oversample=1008, burnin_oversample=1, burnin_time=2000, num_offset=4,
        verbose=False):
        """
        Loads d14c data from a csv file.
        Parameters
        ----------
        file_name : str
            Path to a csv file
        oversample : int, optional
            Number of samples per year in production. 1008 by default
        burnin_oversample : int, optional
            Number of samples per year in burn-in. 1 by default
        burnin_time : int, optional
            Number of years in the burn-in period. 2000 by default
        num_offset : int, optional
            Number of data points used for normalization. 4 by default
        """
        data = Table.read(file_name, format="ascii")
        self.time_data = jnp.array(data["year"])
        self.d14c_data = jnp.array(data["d14c"])
        self.d14c_data_error = jnp.array(data["sig_d14c"])
        self.start = np.nanmin(self.time_data)
        self.end = np.nanmax(self.time_data)
        self.burn_in_time = jnp.arange(self.start - burnin_time, self.start + 1, 1.)
        self.oversample = oversample
        self.burnin_oversample = burnin_oversample
        self.offset = jnp.mean(self.d14c_data[:num_offset])
        self.annual = jnp.arange(self.start, self.end + 1)
        self.mask = jnp.in1d(self.annual, self.time_data)
        self.time_data_fine = jnp.linspace(jnp.min(self.annual), jnp.max(self.annual) + 2,
                                           (self.annual.size + 1) * self.oversample)
        try:
            self.growth = self.get_growth_vector(data["growth_season"][0])
            if verbose:
                with open('data_log.txt', 'a') as f:
                    f.write(" Custom Growth Season")
                    f.write("\n")
        except:
            if self.hemisphere == 'north':
                self.growth = self.get_growth_vector("april-september")
                if verbose:
                    with open('data_log.txt', 'a') as f:
                        f.write(" Growth april-september")
                        f.write("\n")
            else:
                self.growth = self.get_growth_vector("october-march")
                if verbose:
                    with open('data_log.txt', 'a') as f:
                        f.write(" Growth october-march")
                        f.write("\n")

        # define utils for inverse solver now that we have the growth season
        if jnp.count_nonzero(self.growth) == 12:
            def base_interp(time, t_in, data):
                return jnp.interp(time, t_in, data)

            self.interp_type = 'linear'  # keep track of this in case you have to debug
        else:
            def base_interp(time, t_in, data):
                return InterpolatedUnivariateSpline(t_in, data)(time)

            self.interp_type = 'spline'

        interp = jit(base_interp)

        self.dash = jit(grad(interp, argnums=(0)))

    def get_growth_vector(self, growth_season):
        """
        Converts the growing season of a tree from string to 12-digit binary vector.
        Parameters
        ----------
        growth_season : str
            Growing season. Must have the format: "StartMonth-EndMonth"
        Returns
        -------
        ndarray
            12-digit binary vector
        """
        growth_dict = {"january": 0, "february": 1, "march": 2, "april": 3, "may": 4,
                       "june": 5, "july": 6, "august": 7, "september": 8, "october": 9,
                       "november": 10, "december": 11}
        start = growth_dict[growth_season.split('-')[0].lower()]
        end = growth_dict[growth_season.split('-')[1].lower()]
        growth = np.zeros((12,))
        if end < start:
            growth[start:] = 1
            growth[:end + 1] = 1
            self.time_offset = (start / 2 + end / 2 + 6) / 12
        else:
            growth[start:end + 1] = 1
            self.time_offset = (start / 2 + end / 2) / 12
        return jnp.array(growth)

    def compile_production_model(self, model=None):
        """
        Sets the production rate model.
        Parameters
        ----------
        model : str | callable, optional
            Built-in or custom model. Supported built-in model are: "simple_sinusoid", "flexible_sinusoid",
            "flexible_sinusoid_affine_variant", "control_points", "inverse_solver"
        """
        self.production = None
        if callable(model):
            self.production = model
            self.production_model = 'custom'
            self.start_time_index = None
        elif model == "simple_sinusoid":
            self.production = self.simple_sinusoid
            self.production_model = 'simple sinusoid'
            self.start_time_index = 0
        elif model == "spike_only":
            self.production = self.spike_only
            self.production_model = 'spike only'
            self.adaptive = False
            self.start_time_index = 0
        # elif model == "simple_sinusoid_sharp":
        #     self.production = self.simple_sinusoid_sharp
        # elif model == "simple_sinusoid_prolonged":
        #     self.production = self.simple_sinusoid_prolonged
        elif model == "flexible_sinusoid":
            self.production = self.flexible_sinusoid
            self.production_model = 'flexible sinusoid'
            self.start_time_index = 0
        elif model == "flexible_sinusoid_affine_variant":
            self.production = self.flexible_sinusoid_affine_variant
            self.production_model = 'flexible sinusoid affine variant'
            self.start_time_index = 1
        # elif model == "affine":
        #     self.production = self.affine
        #     self.production_model = 'affine'
        elif model == "control_points":
            self.control_points_time = jnp.arange(self.start, self.end)
            self.control_points_time_fine = jnp.linspace(self.start, self.end,
                                                         int((self.end - self.start) * self.oversample))
            self.production = self.interp_gp
            self.production_model = 'control points'
            self.start_time_index = None
        elif model == "inverse_solver":
            self.production = self.interp_IS
            self.production_model = 'inverse solver'
            self.start_time_index = None
        else:
            raise ValueError(
                "model is not a callable, or does not take value from: simple_sinusoid, flexible_sinusoid, "
                "flexible_sinusoid_affine_variant, inverse_solver, control_points")

    @partial(jit, static_argnums=(0,))
    def interp_gp(self, tval, *args):
        """
        A Gaussian Process regression interpolator.
        Parameters
        ----------
        tval : ndarray
            Output time sampling
        args : ndarray | float
            Set of annually resolved control-points passed in as a ndarray
        Returns
        -------
        ndarray
            Interpolation on tval
        """
        tval = tval.reshape(-1)
        params = jnp.array(list(args)).reshape(-1)
        kernel = kernels.Matern32(1)
        gp = GaussianProcess(kernel, self.control_points_time, mean=params[0])
        params = jnp.array(list(args)).reshape(-1)
        return gp.condition(params, tval)[1].loc

    def interp_IS(self, tval, *args):
        """
        A linear interpolator for inverse solver.
        Parameters
        ----------
        tval : ndarray
            Output time sampling
        args : ndarray | float
            Set of production rates on the same time sampling as self.time_data
        Returns
        -------
        ndarray
            Interpolation on tval
        """
        tval = tval.reshape(-1)
        y = jnp.array(list(args)).reshape(-1)
        return jnp.interp(tval, self.time_data, y)

    @partial(jit, static_argnums=(0,))
    def super_gaussian(self, t, start_time, duration, area):
        """
        Computes the density of a super gaussian of exponent 16. Used to emulates the
        spike in d14c data following the occurrence of a Miyake event.
        Parameters
        ----------
        t : ndarray
            Time sampling
        start_time : float
            Start time of a Miyake event
        duration : float
            Duration of a Miyake event
        area : float
            Total radiocarbon delivered by a Miyake event
        Returns
        -------
        ndarray
            Production rate on t
        """
        middle = start_time + duration / 2.
        height = area / duration
        return height * jnp.exp(- ((t - middle) / (1. / 1.93516 * duration)) ** 16.)

    @partial(jit, static_argnums=(0,))
    def simple_sinusoid(self, t, *args):
        """
        A simple sinusoid production rate model. Tunable parameters are,
        Start time: start time\n
        log_Duration: log10 duration\n
        Phase: phase of the solar cycle\n
        log_Area: log10 total radiocarbon delivered
        Parameters
        ----------
        t : ndarray
            Time sampling
        args : ndarray
            Tunable parameters. Must include, start time, duration, phase and area
        Returns
        -------
        ndarray
            Production rate on t
        """
        start_time, log_duration, phase, log_area = jnp.array(list(args)).reshape(-1)
        duration, area = 10**log_duration, 10**log_area
        height = self.super_gaussian(t, start_time, duration, area)
        production = self.steady_state_production + 0.18 * self.steady_state_production * jnp.sin(
            2 * np.pi / 11 * t + phase * 2 * np.pi / 11) + height
        return production


    @partial(jit, static_argnums=(0,))
    def spike_only(self, t, *args):
        """
        A simple sinusoid production rate model. Tunable parameters are,
        Start time: start time\n
        log_Duration: log10 duration\n
        log_Area: total radiocarbon delivered
        Parameters
        ----------
        t : ndarray
            Time sampling
        args : ndarray
            Tunable parameters. Must include, start time, duration, phase and area
        Returns
        -------
        ndarray
            Production rate on t
        """
        start_time, log_duration, log_area = jnp.array(list(args)).reshape(-1)
        duration, area = 10**log_duration, 10**log_area
        height = self.super_gaussian(t, start_time, duration, area)
        production = self.steady_state_production + height
        return production

    @partial(jit, static_argnums=(0,))
    def flexible_sinusoid(self, t, *args):
        """
        A flexible sinusoid production rate model. Tunable parameters are,
        Start time: start time\n
        log_Duration: log10 duration\n
        Phase: phase of the solar cycle\n
        log_Area: log10 total radiocarbon delivered\n
        log_Amplitude: log10 solar amplitude
        Parameters
        ----------
        t : ndarray
            Time sampling
        args : ndarray
            Tunable parameters. Must include, start time, duration, phase, area and amplitude
        Returns
        -------
        ndarray
            Production rate on t
        """
        start_time, log_duration, phase, log_area, log_amplitude = jnp.array(list(args)).reshape(-1)
        duration, area, amplitude = 10**log_duration, 10**log_area, 10**log_amplitude

        height = self.super_gaussian(t, start_time, duration, area)
        production = self.steady_state_production + amplitude * self.steady_state_production * jnp.sin(
            2 * np.pi / 11 * t + phase * 2 * np.pi / 11) + height
        return production



    @partial(jit, static_argnums=(0,))
    def flexible_sinusoid_affine_variant(self, t, *args):
        """
        A flexible sinusoid production rate model with a linear gradient. Tunable parameters are,
        Gradient: linear gradient\n
        Start time: start time\n
        log_Duration: log10 duration\n
        Phase: phase of the solar cycle\n
        log_Area: log10 total radiocarbon delivered\n
        log_Amplitude: log10 solar amplitude
        Parameters
        ----------
        t : ndarray
            Time sampling
        args : ndarray
            Tunable parameters. Must include, gradient, start time, duration, phase, area and amplitude
        Returns
        -------
        ndarray
            Production rate on t
        """
        gradient, start_time, log_duration, phase, log_area, log_amplitude = jnp.array(list(args)).reshape(-1)
        duration, area, amplitude = 10**log_duration, 10**log_area, 10**log_amplitude
        
        height = self.super_gaussian(t, start_time, duration, area)
        production = self.steady_state_production + gradient * (
                t - self.start) * (t >= self.start) + amplitude * self.steady_state_production * jnp.sin(
            2 * np.pi / 11 * t
            + phase * 2 * np.pi / 11) + height
        return production


    # @partial(jit, static_argnums=(0,))
    # def affine(self, t, *args):
    #     gradient, start_time, duration, area = jnp.array(list(args)).reshape(-1)
    #     height = self.super_gaussian(t, start_time, duration, area)
    #     production = self.steady_state_production + gradient * (t - self.start) * (t >= self.start) + height
    #     return production

    @partial(jit, static_argnums=(0))
    def run_burnin(self, y0=None, params=()):
        """
        Calculates the C14 content of all the boxes within a CBM for the burn-in period.
        Parameters
        ----------
        y0 : ndarray, optional
            Initial contents of all boxes
        params : ndarray, optional
            Parameters for the production rate model
        Returns
        -------
        ndarray
            Value of each box in the CBM during the burn-in period
        """
        box_values, _ = self.cbm.run(self.burn_in_time, self.production, y0=y0, args=params,
                                     steady_state_production=self.steady_state_production)
        return box_values

    @partial(jit, static_argnums=(0))
    def run_event(self, y0=None, params=()):
        """
        Calculates the C14 content of all the boxes within a CBM for the production period.
        Parameters
        ----------
        y0 : ndarray, optional
            The initial contents of all boxes
        params : ndarray, optional
            Parameters for self.production function
        Returns
        -------
        ndarray
            Value of each box in the CBM during the production period
        """
        if self.start_time_index is not None:
            start, duration = params[self.start_time_index], 10**params[self.start_time_index+1] # assumes they are adjacent - careful in future
            # also assumes duration is log!
            step_ts = jnp.linspace(start,start+duration,25)
        else:
            step_ts = None
        time_values = jnp.linspace(jnp.min(self.annual), jnp.max(self.annual) + 2,
                                   (self.annual.size + 1) * self.oversample)
        box_values, _ = self.cbm.run(time_values, self.production, y0=y0, args=params,
                                     steady_state_production=self.steady_state_production,
                                     adaptive=self.adaptive,step_ts=step_ts)
        return box_values

    @partial(jit, static_argnums=(0,))
    def dc14(self, params=()):
        """
        Predict d14c on the time sampling from the data file.
        Parameters
        ----------
        params : ndarray, optional
            Parameters for the production rate model
        Returns
        -------
        ndarray
            Predicted d14c value
        """
        if self.adaptive:
            burnin = self.run_burnin(y0=self.steady_state_y0, params=params)
            y0 = burnin[-1, :]
        else:
            y0 = self.steady_state_y0

        event = self.run_event(y0=y0, params=params)

        binned_data = self.cbm.bin_data(event[:, self.box_idx], self.oversample, self.annual, growth=self.growth)
        d14c = (binned_data - self.steady_state_y0[self.box_idx]) / self.steady_state_y0[self.box_idx] * 1000
        return d14c[self.mask] + self.offset

    @partial(jit, static_argnums=(0,))
    def dc14_fine(self, params=()):
        """
        Predict d14c on a sub-annual time sampling.
        Parameters
        ----------
        params : ndarray, optional
            Parameters for the production rate model
        Returns
        -------
        ndarray
            Predicted d14c value
        """
        if self.adaptive:
            burnin = self.run_burnin(y0=self.steady_state_y0, params=params)
            y0 = burnin[-1, :]
        else:
            y0 = self.steady_state_y0

        event = self.run_event(y0=y0, params=params)
        d14c = (event[:, self.box_idx] - self.steady_state_y0[self.box_idx]) / self.steady_state_y0[self.box_idx] * 1000
        return d14c + self.offset

    # @partial(jit, static_argnums=(0,))
    def log_likelihood(self, params=()):
        """
        Computes the gaussian log-likelihood of production rate model parameters.
        Parameters
        ----------
        params : ndarray, optional
            Parameters of the production rate model
        Returns
        -------
        float
            Gaussian log-likelihood
        """
        d14c = self.dc14(params)
        return -0.5 * jnp.sum(((self.d14c_data - d14c) / self.d14c_data_error) ** 2)

    @partial(jit, static_argnums=(0,))
    def log_joint_likelihood(self, params, low_bounds, up_bounds):
        """
        Computes the log joint likelihood of production rate model parameters.
        Parameters
        ----------
        params : ndarray
            Production rate model parameters
        low_bounds : ndarray
            Lower bound of params
        up_bounds : ndarray
            Upper bound of params
        Returns
        -------
        float
            Log joint likelihood
        """
        lp = 0
        lp += jnp.any((params < low_bounds) | (params > up_bounds)) * -jnp.inf
        pos = self.log_likelihood(params)
        return lp + pos


    @partial(jit, static_argnums=(0,))
    def log_likelihood_gp(self, params):
        """
        Computes the Gaussian Process log-likelihood of a set of control-points. The Gaussian Process
        has a constant mean and a Matern-3/2 kernel with 1 year scale parameter.
        Parameters
        ----------
        params : ndarray
            An array of control-points. The first control point is also the mean of the Gaussian Process
        Returns
        -------
        float
            Gaussian Process log-likelihood
        """
        kernel = kernels.Matern32(1)
        gp = GaussianProcess(kernel, self.control_points_time, mean=params[0])
        return gp.log_probability(params)

    @partial(jit, static_argnums=(0,))
    def log_joint_likelihood_gp(self, params, low_bounds, up_bounds):
        """
        Computes the log joint likelihood of a set of control-points.
        Parameters
        ----------
        params : ndarray
            An array of control-points
        Returns
        -------
        float
            Log joint likelihood
        """
        lp = jnp.any((params < low_bounds) | (params > up_bounds)) * -jnp.inf
        return self.log_likelihood(params=params) + self.log_likelihood_gp(params) + lp

    @partial(jit, static_argnums=(0,))
    def neg_log_joint_likelihood_gp(self, params):
        """
        Computes the negative log joint likelihood of a set of control-points. Used as the objective function for
        fitting the set of control-points via numerical optimization.
        Parameters
        ----------
        params : ndarray
            An array of control-points
        Returns
        -------
        float
            Negative log joint likelihood
        """
        return -1 * self.log_likelihood(params=params) + -1 * self.log_likelihood_gp(params)

    @partial(jit, static_argnums=(0,))
    def grad_neg_log_joint_likelihood_gp(self, params=()):
        """
        Computes the negative gradient of the log joint likelihood of a set of control-points.
        Parameters
        ----------
        params : ndarray
            An array of control-points
        Returns
        -------
        float
            Negative gradient of log joint likelihood
        """
        return grad(self.neg_log_joint_likelihood_gp)(params)

    def fit_ControlPoints(self, low_bound=0):
        """
        Fits control-points by minimizing the negative log joint likelihood.
        Parameters
        ----------
        low_bound : int, optional
            The minimum value each control-point can take. 0 by default.
        Returns
        -------
        OptimizeResult
            Scipy OptimizeResult object
        """
        initial = self.steady_state_production * jnp.ones((len(self.control_points_time),))
        bounds = tuple([(low_bound, None)] * len(initial))
        soln = scipy.optimize.minimize(self.neg_log_joint_likelihood_gp, initial, bounds=bounds,
                                       options={'maxiter': 100000, 'maxfun': 100000, })
        return soln

    @partial(jit, static_argnums=(0))
    def _reverse_convert_production_rate(self, production_rate):
        new_rate = None
        if self.cbm._production_rate_units == 'atoms/cm^2/s':
            new_rate = production_rate / (14.003242 / 6.022 * 5.11 * 31536. / 1.e5)
        elif self.cbm._production_rate_units == 'kg/yr':
            new_rate = production_rate
        return new_rate

    @partial(jit, static_argnums=(0, 5, 6))
    def reconstruct_production_rate(self, d14c, t_in, t_out, steady_state_solution, steady_state_production=None,
                                    target_C_14=None):
        """
        Parameters
        ----------
        d14c
        t_in
        t_out
        steady_state_solution
        steady_state_production
        target_C_14
        Returns
        -------
        """

        data = d14c / 1000 * steady_state_solution[self.box_idx] + steady_state_solution[self.box_idx]
        first1 = jnp.where(self.growth == 1, size=1)[0][0]
        first0 = jnp.where(self.growth == 0, size=1)[0][0]
        all1s = jnp.where(self.growth == 1, size=12)[0]
        after1 = jnp.where(all1s > first0, all1s, 0)
        after1 = after1.at[jnp.nonzero(after1, size=1)].get()[0]
        num = sub(first1, after1)
        val = cond(num == 0, lambda x: first1, lambda x: after1, num)
        act = cond(jnp.all(self.growth == 1), lambda x: 0, lambda x: val, self.growth)
        act = act + jnp.count_nonzero(self.growth) / 2
        t_in = t_in + act / 12

        dash = lambda x: self.dash(x, t_in, data)

        # @jit
        def derivative(time, y, args):
            ans = jnp.matmul(self.cbm.get_matrix(), y)
            prod_coeff = self.cbm.get_production_coefficients()
            production_rate = (dash(time) - ans[self.box_idx]) / prod_coeff[self.box_idx]
            production_term = prod_coeff * production_rate
            return ans + production_term

        if target_C_14 is not None:
            steady_state = self.cbm.equilibrate(production_rate=self.cbm.equilibrate(target_C_14=target_C_14))
        elif steady_state_production is not None:
            steady_state = self.cbm.equilibrate(production_rate=steady_state_production)
        else:
            raise ValueError("Must give either target C-14 or production rate.")

        term = diffrax.ODETerm(derivative)
        solver = diffrax.Dopri5()
        saveat = diffrax.SaveAt(ts=t_out)

        # for fixed time sampling
        step = 1 / 48. # 4 per month 
        max_steps = 96*2*jnp.size(t_out)
        
        stepsize_controller = diffrax.PIDController(rtol=1e-10, atol=1e-10)

        states = diffrax.diffeqsolve(term, solver, args=(), y0=steady_state, t0 = t_out[0], t1 = t_out[-1],
            dt0 = step, stepsize_controller=stepsize_controller,saveat=saveat)

        flows = jnp.matmul(self.cbm.get_matrix(), states.ys.T)
        return self._reverse_convert_production_rate((vmap(dash)(t_out) - flows[self.box_idx, :]) /
                                                     self.cbm.get_production_coefficients()[self.box_idx])

    def MC_reconstruct(self, iters=1000, t_in=None, t_out=None):
        """
        Parameters
        ----------
        iters : int, optional
            number of iters to do the chain of.
        t_in
        t_out
        Returns
        -------
        """
        if t_in is None:
            t_in = jnp.zeros((self.time_data.size + 1))
            t_in = t_in.at[jnp.arange(self.time_data.size + 1)[1:]].set(self.time_data)
            t_in = t_in.at[0].set(self.start - 1)
            # t_in = jax.ops.index_update(t_in, jnp.arange(self.time_data.size + 1)[1:], self.time_data)
            # t_in = jax.ops.index_update(t_in, 0, self.start - 1)

        if t_out is None:
            t_out = self.time_data.astype('float64')

        production_rates = []

        for _ in tqdm(range(iters)):
            new_data = self.d14c_data - self.offset + np.random.randn(len(self.d14c_data)) * self.d14c_data_error
            new_data = np.concatenate((jnp.expand_dims(jnp.array(0), axis=0), new_data))
            prod_recon = self.reconstruct_production_rate(new_data, t_in, t_out, self.steady_state_y0,
                                                          steady_state_production=self.steady_state_production)
            production_rates.append(prod_recon)

        chain = np.array(production_rates)
        return chain


class MultiFitter(CarbonFitter):
    """
    A class for parametric inference of d14c data from a common time period using an ensemble of SingleFitter.
    Does parameter fitting, likelihood evaluations, Monte Carlo sampling, plotting and more.
    """

    def __init__(self,adaptive=True):
        """
        Initializes a MultiFitter object.
        """
        self.MultiFitter = []
        self.burnin_oversample = 1
        self.start = None
        self.end = None
        self.oversample = None
        self.production = None
        self.production_model = None
        self.burn_in_time = None
        self.steady_state_y0 = None
        self.steady_state_production = None
        self.growth = None
        self.cbm = None
        self.cbm_model = None
        self.box_idx = None
        self.adaptive=adaptive

    def add_SingleFitter(self, sf):
        """
        Adds a SingleFitter object to a Multifitter object.
        Parameters
        ----------
        sf : SingleFitter
            SingleFitter Object
        """
        if not self.start:
            self.start = sf.start
        elif self.start > sf.start:
            self.start = sf.start

        if not self.end:
            self.end = sf.end
        elif self.end < sf.end:
            self.end = sf.end

        if self.production is None:
            self.production = sf.production
            self.production_model = sf.production_model
            self.start_time_index = sf.start_time_index
        elif self.production_model is not sf.production_model:
            raise ValueError(
                "production for SingleFitters must be consistent. Got {}, expected {}".format(sf.production_model,
                                                                                              self.production_model))
        if self.oversample is None:
            self.oversample = sf.oversample
        elif self.oversample < sf.oversample:
            self.oversample = sf.oversample

        # if not self.box_idx:
        #     self.box_idx = sf.box_idx
        # elif not self.box_idx == sf.box_idx:
        #     raise ValueError(
        #         "'box' parameter and 'hemisphere' parameter for SingleFitters must be consistent")

        if self.steady_state_y0 is None:
            self.steady_state_y0 = sf.steady_state_y0
            self.steady_state_production = sf.steady_state_production
        elif not jnp.allclose(self.steady_state_y0, sf.steady_state_y0):
            raise ValueError(
                "steady state burn-in solution for SingleFitters must be consistent. Got {}, expected {}".format(
                    sf.steady_state_y0,
                    self.steady_state_y0))

        if self.cbm is None:
            self.cbm_model = sf.cbm_model
            self.cbm = ticktack.load_presaved_model(self.cbm_model, production_rate_units='atoms/cm^2/s')
            self.cbm.compile()
        # elif not self.cbm_model is sf.cbm_model:
        #     raise ValueError("cbm model for SingleFitters must be consistent. Got {}, expected {}".format(sf.cbm_model,
        #                                                                                                   self.cbm_model))
        self.MultiFitter.append(sf)

    def compile(self):
        """
        Prepares a Multifitter object for d14c computation and likelihood evaluation.
        """
        if self.production_model == 'flexible sinusoid affine variant':
            self.production = self.flexible_sinusoid_affine_variant
        self.burn_in_time = jnp.arange(self.start - 2000, self.start + 1, 1.)
        self.annual = jnp.arange(self.start, self.end + 1)
        self.time_data_fine = jnp.linspace(jnp.min(self.annual), jnp.max(self.annual) + 2,
                                           (self.annual.size + 1) * self.oversample)
        for sf in self.MultiFitter:
            sf.multi_mask = jnp.in1d(self.annual, sf.time_data)
        if self.production_model == 'control points':
            self.control_points_time = jnp.arange(self.start, self.end)
            self.production = self.multi_interp_gp
        self.steady_state_box = self.steady_state_y0[self.box_idx]

    @partial(jit, static_argnums=(0,))
    def multi_interp_gp(self, tval, *args):
        """
        A Gaussian Process regression interpolator for MultiFitter.
        Parameters
        ----------
        tval : ndarray
            Output time sampling
        args : ndarray | float
            Set of annually resolved control-points
        Returns
        -------
        ndarray
            Interpolated values on tval
        """
        tval = tval.reshape(-1)
        params = jnp.array(list(args)).reshape(-1)
        kernel = kernels.Matern32(1)
        gp = GaussianProcess(kernel, self.control_points_time, mean=params[0])
        params = jnp.array(list(args)).reshape(-1)
        return gp.condition(params, tval)[1].loc

    @partial(jit, static_argnums=(0,))
    def super_gaussian(self, t, start_time, duration, area):
        """
        Computes the density of a super gaussian of exponent 16. Used to emulates the
        spike in d14c data following the occurrence of a Miyake event.
        Parameters
        ----------
        t : ndarray
            Time sampling
        start_time : float
            Start time of a Miyake event
        duration : float
            Duration of a Miyake event
        area : float
            Total radiocarbon delivered by a Miyake event
        Returns
        -------
        ndarray
            Production rate on t
        """
        middle = start_time + duration / 2.
        height = area / duration
        return height * jnp.exp(- ((t - middle) / (1. / 1.93516 * duration)) ** 16.)

    @partial(jit, static_argnums=(0,))
    def flexible_sinusoid_affine_variant(self, t, *args):
        """
        A flexible sinusoid production rate model with a linear gradient. Tunable parameters are,
        Gradient: linear gradient\n
        Start time: start time\n
        log_Duration: log10 duration\n
        Phase: phase of the solar cycle\n
        log_Area: log10 total radiocarbon delivered\n
        log_Amplitude: log10 solar amplitude
        Parameters
        ----------
        t : ndarray
            Time sampling
        args : ndarray
            Tunable parameters. Must include, gradient, start time, duration, phase, area and amplitude
        Returns
        -------
        ndarray
            Production rate on t
        """
        gradient, start_time, log_duration, phase, log_area, log_amplitude = jnp.array(list(args)).reshape(-1)
        duration, area, amplitude = 10**log_duration, 10**log_area, 10**log_amplitude

        height = self.super_gaussian(t, start_time, duration, area)
        production = self.steady_state_production + gradient * (
                t - self.start) * (t >= self.start) + amplitude * self.steady_state_production * jnp.sin(
            2 * np.pi / 11 * t
            + phase * 2 * np.pi / 11) + height
        return production

    @partial(jit, static_argnums=(0))
    def run_burnin(self, y0=None, params=()):
        """
        Calculates the C14 content of all the boxes within a CBM for the burn-in period.
        Parameters
        ----------
        y0 : ndarray, optional
            Initial contents of all boxes
        params : ndarray, optional
            Parameters for the production rate model
        Returns
        -------
        ndarray
            Value of each box in the CBM during the burn-in period
        """
        box_values, _ = self.cbm.run(self.burn_in_time, self.production, y0=y0, args=params,
                                     steady_state_production=self.steady_state_production)
        return box_values

    @partial(jit, static_argnums=(0))
    def run_event(self, y0=None, params=()):
        """
        Calculates the C14 content of all the boxes within a CBM for the production period.
        Parameters
        ----------
        y0 : ndarray, optional
            The initial contents of all boxes
        params : ndarray, optional
            Parameters for self.production function
        Returns
        -------
        ndarray
            Value of each box in the CBM during the production period
        """
        if self.start_time_index is not None:
            start, duration = params[self.start_time_index], 10**params[self.start_time_index+1] # assumes they are adjacent - careful in future
            # assumes duration is log space
            step_ts = jnp.linspace(start,start+duration,25)
        else:
            step_ts = None
        time_values = jnp.linspace(jnp.min(self.annual), jnp.max(self.annual) + 2,
                                   (self.annual.size + 1) * self.oversample)
        box_values, _ = self.cbm.run(time_values, self.production, y0=y0, args=params,
                                     steady_state_production=self.steady_state_production,
                                     adaptive=self.adaptive)
        return box_values

    @partial(jit, static_argnums=(0,))
    def dc14_fine(self, params=()):
        """
               Predict d14c on a sub-annual time sampling.
               Parameters
               ----------
               params : ndarray, optional
                   Parameters for the production rate model
               Returns
               -------
               ndarray
                   Predicted d14c value
               """
        if self.adaptive:
            burnin = self.run_burnin(y0=self.steady_state_y0, params=params)
            y0 = burnin[-1, :]
        else:
            y0 = self.steady_state_y0
        event = self.run_event(y0=y0, params=params)
        d14c = (event[:, self.box_idx] - self.steady_state_y0[self.box_idx]) / self.steady_state_y0[self.box_idx] * 1000
        return d14c

    # @partial(jit, static_argnums=(0,))
    def multi_likelihood(self, params):
        """
        Computes the ensemble log-likelihood of the parameters of self.production across multiple d14c datasets
        Parameters
        ----------
        params : ndarray
            Parameters of self.production
        Returns
        -------
        float
            Log-likelihood
        """
        if self.adaptive:
            burnin = self.run_burnin(y0=self.steady_state_y0, params=params) # steady_state_y0 is hem independent
            y0 = burnin[-1, :]
        else:
            y0 = self.steady_state_y0

        event_mat = self.run_event(y0=y0, params=params) # gives a list of boxes
        like = 0
        for sf in self.MultiFitter:
            event = event_mat[:, sf.box_idx] # chooses the right box using the information from sf
            binned_data = sf.cbm.bin_data(event, self.oversample, self.annual, growth=sf.growth) # bins using sf.cbm
            d14c = (binned_data - sf.steady_state_box) / sf.steady_state_box * 1000 # steady_state_box is hem dependent
            d14c_sf = d14c[sf.multi_mask] + sf.offset
            like += jnp.sum(((sf.d14c_data - d14c_sf) / sf.d14c_data_error) ** 2) * -0.5
        return like

    @partial(jit, static_argnums=(0,))
    def log_likelihood_gp(self, params):
        """
        Computes the Gaussian Process log-likelihood of a set of control-points. The Gaussian Process
        has a constant mean and a Matern-3/2 kernel with 1 year scale parameter.
        Parameters
        ----------
        params : ndarray
            An array of control-points. The first control point is also the mean of the Gaussian Process
        Returns
        -------
        float
            Gaussian Process log-likelihood
        """
        kernel = kernels.Matern32(1)
        gp = GaussianProcess(kernel, self.control_points_time, mean=params[0])
        return gp.log_probability(params)

    def log_joint_likelihood(self, params, low_bounds, up_bounds):
        """
        Computes the log joint likelihood of production rate model parameters.
        Parameters
        ----------
        params : ndarray
            Production rate model parameters
        low_bounds : ndarray
            Lower bound of params
        up_bounds : ndarray
            Upper bound of params
        Returns
        -------
        float
            Log joint likelihood
        """
        lp = 0
        lp += jnp.any((params < low_bounds) | (params > up_bounds)) * -jnp.inf
        pos = self.multi_likelihood(params)
        return lp + pos


    def log_joint_likelihood_gp(self, params, low_bounds, up_bounds):
        """
        Computes the log joint likelihood of a set of control-points.
        Parameters
        ----------
        params : ndarray
            An array of control-points
        Returns
        -------
        float
            Log joint likelihood
        """
        lp = jnp.any((params < low_bounds) | (params > up_bounds)) * -jnp.inf
        return self.multi_likelihood(params=params) + self.log_likelihood_gp(params) + lp

    def neg_log_joint_likelihood_gp(self, params):
        """
        Computes the negative log joint likelihood of a set of control-points. Used as the objective function for
        fitting the set of control-points via numerical optimization.
        Parameters
        ----------
        params : ndarray
            An array of control-points
        Returns
        -------
        float
            Negative log joint likelihood
        """
        return -1 * self.multi_likelihood(params=params) + -1 * self.log_likelihood_gp(params)

    def fit_ControlPoints(self, low_bound=0):
        """
        Fits the control-points by minimizing the negative log joint likelihood.
        Parameters
        ----------
        low_bound : int, optional
            The minimum value each control-point can take. 0 by default.
        Returns
        -------
        OptimizeResult
            Scipy OptimizeResult object
        """
        initial = self.steady_state_production * jnp.ones((len(self.control_points_time),))
        bounds = tuple([(low_bound, None)] * len(initial))
        soln = scipy.optimize.minimize(self.neg_log_joint_likelihood_gp, initial, bounds=bounds,
                                       options={'maxiter': 100000, 'maxfun': 100000, })
        return soln.x


def get_data(path=None):
    """
    Retrieves the names of all data files in a directory
    Parameters
    ----------
    path : str, optional
        Path to a directory where data files are stored
    Returns
    -------
    list
        A list of file names
    """
    if path:
        file_names = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]
    elif event in ['660BCE-Ew', '660BCE-Lw', '775AD-early-N', '775AD-early-S', '775AD-late-N',
                   '993AD-N', '993AD-S', '5259BCE', '5410BCE', '7176BCE']:
        file = 'data/datasets/' + event
        path = os.path.join(os.path.dirname(__file__), file)
        file_names = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]
    else:
        raise ValueError(
            "Invalid path, or event is not from the following: '660BCE-Ew', '660BCE-Lw', '775AD-early', '775AD-late', '993AD', '5259BCE', '5410BCE', '7176BCE'")
    return file_names


def sample_event(year, mf, sampler='MCMC', production_model='simple_sinusoid', burnin=500, production=1000,
                 params=None, low_bounds=None, up_bounds=None):
    """
    Runs a Monte Carlo sampler on some data files.
    
    Parameters
    ----------
    year : float
        The calender year of the event
    mf : MultiFitter
        A compiled MultiFitter
    sampler : str, optional
        Monte Carlo sampler. 'MCMC' for Markov Chain Monte Carlo, 'NS' for Nested Sampling. 'MCMC' by default.
    production_model : str | callable, optional
        Production rate model. 'simple_sinusoid' by default.
    burnin : int, optional
        Number of burn-in steps for Markov Chain Monte Carlo, 500 by default.
    production : int, optional
        Number of production steps for Markov Chain Monte Carlo, 1000 by default.
    params : ndarray, optional
        Initial parameters for Monte Carlo samplers. Required when custom production rate model is used.
    low_bounds : ndarray, optional
        Lower bound of params. Required when custom production rate model is used.
    up_bounds : ndarray, optional
        Upper bound of params. Required when custom production rate model is used.
        
    Returns
    -------
    ndarray
        Monte Carlo samples
    """
    if production_model == 'simple_sinusoid':
        default_params = jnp.array([year, np.log10(2. / 12), 3., np.log10(81. / 12)])
        default_low_bounds = jnp.array([year - 5, np.log10(4 / 52.), 0, -2.])
        default_up_bounds = jnp.array([year + 5, np.log10(5.), 11, 1.5])
    elif production_model == 'flexible_sinusoid':
        default_params = jnp.array([year, np.log10(2. / 12), 3., np.log10(81. / 12), np.log10(0.18)])
        default_low_bounds = jnp.array([year - 5, np.log10(4 / 52.), 0, -2, -2])
        default_up_bounds = jnp.array([year + 5, np.log10(5.), 11, 1.5, 1.5])
    elif production_model == 'flexible_sinusoid_affine_variant':
        default_params = jnp.array([0, year, np.log10(2. / 12), 3., np.log10(81. / 12), np.log10(0.18)])
        default_low_bounds = jnp.array([-mf.steady_state_production * 0.05 / 5, year - 5, np.log10(4 / 52.), 0, -2, -2])
        default_up_bounds = jnp.array([mf.steady_state_production * 0.05 / 5, year + 5, np.log10(5.), 11, 1.5, 1.5])
    # elif production_model == 'affine':
    #     default_params = jnp.array([0, year, np.log10(1. / 12), np.log10(81. / 12)])
    #     default_low_bounds = jnp.array([-mf.steady_state_production * 0.05 / 5, year - 5, np.log10(1 / 52.), -2.])
    #     default_up_bounds = jnp.array([mf.steady_state_production * 0.05 / 5, year + 5, 2., 1.5])
    elif production_model == 'control_points':
        if sampler == "MCMC":
            default_params = mf.steady_state_production * jnp.ones((len(mf.control_points_time),))
            default_low_bounds = jnp.array([0] * default_params.size)
            default_up_bounds = jnp.array([100] * default_params.size)
        elif sampler == "optimisation":
            print("Running numerical optimization...")
            return mf, mf.fit_ControlPoints()
    if all(arg is not None for arg in (low_bounds, up_bounds)):
        low_bounds = low_bounds
        up_bounds = up_bounds
    else:
        low_bounds = default_low_bounds
        up_bounds = default_up_bounds

    if params is not None:
        params = params
    else:
        params = default_params

    if sampler == 'MCMC':
        if production_model != 'control_points':
            chain = mf.MarkovChainSampler(params,
                                          likelihood=mf.log_joint_likelihood,
                                          burnin=burnin,
                                          production=production,
                                          args=(low_bounds, up_bounds)
                                          )
        else:
            chain = mf.MarkovChainSampler(params,
                                          likelihood=mf.log_joint_likelihood_gp,
                                          burnin=burnin,
                                          production=production,
                                          args=(low_bounds, up_bounds))
    # elif sampler == 'NS':
    #     print("Running Nested Sampling...")
    #     chain = mf.NestedSampler(params,
    #                              likelihood=mf.multi_likelihood,
    #                              low_bound=low_bounds,
    #                              high_bound=up_bounds
    #                              )
    #     print("Done")
    else:
        raise ValueError("Invalid sampler value. 'sampler' must be one of the following: MCMC, NS")
    return chain


def fit_event(year, path=None, production_model='simple_sinusoid', cbm_model='Guttler15', box='Troposphere',
              hemisphere='north', sampler=None, burnin=500, production=1000, params=None, low_bounds=None,
              up_bounds=None, mf=None, oversample=1008, burnin_time=2000, verbose=False):
    """
    Fits a Miyake event.

    Parameters
    ----------
    year : float
        The calender year of the event
    mf : MultiFitter, optional
        A compiled MultiFitter
    cbm_model : str, optional
        Name of a Carbon Box Model. Must be one in: Miyake17, Brehm21, Guttler15, Buntgen18.
    oversample : int, optional
        Number of samples per year in production. 1008 by default
    burnin_time : int, optional
        Number of years in the burn-in period. 2000 by default
    sampler : str, optional
        Monte Carlo sampler. 'MCMC' for Markov Chain Monte Carlo, 'NS' for Nested Sampling. 'MCMC' by default.
    box : str, optional
        Box for calculating d14c. 'Troposphere' by default
    hemisphere : str, optional
        CBM hemisphere. Must be one in: "north", "south". "north" by default
    production_model : str | callable, optional
        Production rate model. 'simple_sinusoid' by default.
    burnin : int, optional
        Number of burn-in steps for Markov Chain Monte Carlo, 500 by default.
    production : int, optional
        Number of production steps for Markov Chain Monte Carlo, 1000 by default.
    params : ndarray, optional
        Initial parameters for Monte Carlo samplers. Required when custom production rate model is used.
    low_bounds : ndarray, optional
        Lower bound of params. Required when custom production rate model is used.
    up_bounds : ndarray, optional
        Upper bound of params. Required when custom production rate model is used.
        
    Returns
    -------
    mf : MultiFitter
        A compile MultiFitter object
    chain : ndrray | NestedSampler object
        Monte Carlo samples
    """
    if not mf:
        mf = MultiFitter()
    cbm = ticktack.load_presaved_model(cbm_model, production_rate_units='atoms/cm^2/s')
    file_names = get_data(path=path)
    print("Retrieving data...")
    for file_name in tqdm(file_names):
        if 'SH' in file_name:
            sf = SingleFitter(cbm, cbm_model, box=box, hemisphere='south')
            if verbose:
                with open('data_log.txt', 'a') as f:
                    f.write("%s: Hemisphere South," % file_name)
        else:
            sf = SingleFitter(cbm, cbm_model, box=box, hemisphere=hemisphere)
            if verbose:
                with open('data_log.txt', 'a') as f:
                    f.write("%s: Hemisphere North," % file_name)
        sf.load_data(path + '/' + file_name, oversample=oversample, burnin_time=burnin_time, verbose=verbose)
        sf.compile_production_model(model=production_model)
        mf.add_SingleFitter(sf)
    mf.compile()
    if not sampler:
        return mf
    else:
        return mf, sample_event(year, mf, sampler=sampler, params=params, burnin=burnin, production=production,
                                production_model=production_model, low_bounds=low_bounds, up_bounds=up_bounds)


def plot_samples(average_path=None, chains_path=None, cbm_models=None, cbm_label=None, hemisphere="north",
                 production_model=None,
                 directory_path=None, size=50, size2=30, alpha=0.05, alpha2=0.1, savefig_path=None, title=None,
                 axs=None, labels=True, interval=None, capsize=3, markersize=6, elinewidth=3):
    if axs:
        ax1, ax2 = axs
    else:
        fig, (ax1, ax2) = plt.subplots(2, figsize=(8, 8), sharex=True, gridspec_kw={'height_ratios': [2, 1]})
        fig.subplots_adjust(hspace=0.05)
    colors = mpl.rcParams['axes.prop_cycle'].by_key()['color']
    for i, model in enumerate(cbm_models):
        cbm = ticktack.load_presaved_model(model, production_rate_units='atoms/cm^2/s')
        sf = SingleFitter(cbm, cbm_model=model, hemisphere=hemisphere)
        sf.load_data(average_path)
        chain = np.load(chains_path[i])
        sf.compile_production_model(model=production_model)

        if sf.start < 0:
            time_data = sf.time_data * -1 - sf.time_offset
            time_data_fine = sf.time_data_fine * -1
            ax1.invert_xaxis()
            ax2.invert_xaxis()
        else:
            time_data = sf.time_data + sf.time_offset
            time_data_fine = sf.time_data_fine

        idx = np.random.randint(len(chain), size=size)
        for param in chain[idx]:
            ax1.plot(time_data_fine, sf.dc14_fine(params=param), alpha=alpha, color=colors[i])

        for param in chain[idx][:size2]:
            ax2.plot(time_data_fine, sf.production(sf.time_data_fine, *param), alpha=alpha2, color=colors[i])

    ax1.errorbar(time_data, sf.d14c_data, yerr=sf.d14c_data_error, fmt="ok", capsize=capsize, markersize=markersize,
                 elinewidth=elinewidth, label=r"average $\Delta^{14}$C")

    if directory_path:
        file_names = [f for f in os.listdir(directory_path) if os.path.isfile(os.path.join(directory_path, f))]
        for file in file_names:
            sf.load_data(directory_path + '/' + file)
            if sf.start < 0:
                ax1.errorbar(-sf.time_data - sf.time_offset, sf.d14c_data, fmt="o", color="gray",
                             yerr=sf.d14c_data_error, capsize=capsize,
                             markersize=markersize, alpha=0.2, elinewidth=elinewidth)
            else:
                ax1.errorbar(sf.time_data + sf.time_offset, sf.d14c_data, fmt="o", color="gray",
                             yerr=sf.d14c_data_error, capsize=capsize,
                             markersize=markersize, alpha=0.2, elinewidth=elinewidth)
    sf.load_data(average_path)
    if labels:
        if cbm_label:
            custom_lines = [Line2D([0], [0], color=colors[i], lw=1.5, label=cbm_label[i]) for i in
                            range(len(cbm_label))]
        else:
            custom_lines = [Line2D([0], [0], color=colors[i], lw=1.5, label=cbm_models[i]) for i in
                            range(len(cbm_models))]
        custom_lines.append(Line2D([0], [0], color="k", marker="o", lw=1.5, label=r"average $\Delta^{14}$C"))
        ax1.legend(handles=custom_lines)
        ax1.set_ylabel(r"$\Delta^{14}$C (‰)")
        if sf.start < 0:
            ax2.set_xlabel("Year (BCE)");
        else:
            ax2.set_xlabel("Year (AD)");
        ax2.set_ylabel("Production rate (atoms/cm$^2$/s)");
        plt.suptitle(title);
        plt.tight_layout();
    ax2.set_ylim(1, 10);
    if interval:
        if sf.start < 0:
            ax2.set_xlim(-sf.start + 0.2, -sf.end - 0.2);
            ax2.set_xticks(np.arange(-sf.end - 1, -sf.start, interval))
        else:
            ax2.set_xlim(sf.start - 0.2, sf.end + 0.2);
            ax2.set_xticks(np.arange(sf.start, sf.end + 1, interval))
    else:
        if sf.start < 0:
            ax2.set_xlim(-sf.start + 0.2, -sf.end - 0.2);
        else:
            ax2.set_xlim(sf.start - 0.2, sf.end + 0.2);
        ax2.xaxis.set_major_locator(MaxNLocator(integer=True))

    if savefig_path:
        plt.savefig(savefig_path)


def plot_ControlPoints(average_path=None, soln_path=None, chain_path=None, cbm_models=None, cbm_label=None,
                       hemisphere="north", merged_inverse_solver=None,
                       directory_path=None, savefig_path=None, title=None, axs=None, labels=True, interval=None,
                       markersize=6, capsize=3, markersize2=3, elinewidth=3, size=1, alpha=1, ):
    if axs:
        ax1, ax2 = axs
    else:
        fig, (ax1, ax2) = plt.subplots(2, dpi=100, figsize=(8, 8), sharex=True, gridspec_kw={'height_ratios': [2, 1]})
        fig.subplots_adjust(hspace=0.05)
    colors = mpl.rcParams['axes.prop_cycle'].by_key()['color']
    for i, model in enumerate(cbm_models):
        cbm = ticktack.load_presaved_model(model, production_rate_units='atoms/cm^2/s')
        sf = SingleFitter(cbm, cbm_model=model, hemisphere=hemisphere)
        sf.load_data(average_path)
        soln = np.load(soln_path[i], allow_pickle=True)
        sf.compile_production_model(model="control_points")

        if sf.start < 0:
            time_data = sf.time_data * -1 - sf.time_offset
            time_data_fine = sf.time_data_fine * -1
            control_points_time = sf.control_points_time * -1
            control_points_time_fine = sf.control_points_time_fine * -1
            ax1.invert_xaxis()
            ax2.invert_xaxis()
        else:
            time_data = sf.time_data + sf.time_offset
            control_points_time = sf.control_points_time
            control_points_time_fine = sf.control_points_time_fine
            time_data_fine = sf.time_data_fine

        if np.all(merged_inverse_solver is not None):
            ax2.errorbar(time_data, np.median(merged_inverse_solver, axis=0), fmt="k", drawstyle="steps",
                         alpha=0.2)
            ax2.fill_between(time_data, np.percentile(merged_inverse_solver, 32, axis=0),
                             np.percentile(merged_inverse_solver, 68, axis=0), step='pre', alpha=0.1,
                             color="k", edgecolor="none", lw=1.5)

        if chain_path:
            chain = np.load(chain_path[i], allow_pickle=True)
            mu = np.mean(chain, axis=0)

            if size == 1:
                ax1.plot(time_data_fine, sf.dc14_fine(mu), color=colors[i])
            else:
                idx = np.random.randint(len(chain), size=size)
                for param in chain[idx]:
                    ax1.plot(time_data_fine, sf.dc14_fine(params=param), alpha=alpha, color=colors[i])

            ax2.plot(control_points_time_fine, sf.interp_gp(sf.control_points_time_fine, mu), color=colors[i])
            idx = np.random.randint(len(chain), size=30)
            for param in chain[idx]:
                ax2.plot(control_points_time_fine, sf.interp_gp(sf.control_points_time_fine, param),
                         alpha=0.2, color=colors[i])
        else:
            ax1.plot(time_data_fine, sf.dc14_fine(soln), color=colors[i])
            ax2.plot(control_points_time_fine, sf.interp_gp(sf.control_points_time_fine, soln), color=colors[i])

    ax1.errorbar(time_data, sf.d14c_data, yerr=sf.d14c_data_error, fmt="ok", capsize=capsize,
                 markersize=markersize, elinewidth=elinewidth, label=r"average $\Delta^{14}$C")

    if directory_path:
        file_names = [f for f in os.listdir(directory_path) if os.path.isfile(os.path.join(directory_path, f))]
        for file in file_names:
            sf.load_data(directory_path + '/' + file)
            if sf.start < 0:
                ax1.errorbar(-sf.time_data - sf.time_offset, sf.d14c_data, fmt="o", color="gray",
                             yerr=sf.d14c_data_error, capsize=capsize,
                             markersize=markersize, alpha=0.2)
            else:
                ax1.errorbar(sf.time_data + sf.time_offset, sf.d14c_data, fmt="o", color="gray",
                             yerr=sf.d14c_data_error, capsize=capsize,
                             markersize=markersize, alpha=0.2)

    sf.load_data(average_path)
    if labels:
        if cbm_label:
            custom_lines = [Line2D([0], [0], color=colors[i], lw=1.5, label=cbm_label[i]) for i in
                            range(len(cbm_label))]
        else:
            custom_lines = [Line2D([0], [0], color=colors[i], lw=1.5, label=cbm_models[i]) for i in
                            range(len(cbm_models))]
        custom_lines.append(Line2D([0], [0], color="k", marker="o", lw=1.5, label=r"average $\Delta^{14}$C"))
        ax1.legend(handles=custom_lines, frameon=False)
        ax1.set_ylabel(r"$\Delta^{14}$C (‰)")
        if sf.start < 0:
            ax2.set_xlabel("Year (BCE)");
        else:
            ax2.set_xlabel("Year (AD)");
        ax2.set_ylabel("Production rate (atoms/cm$^2$/s)");
        plt.suptitle(title);
        plt.tight_layout();
    if interval:
        if sf.start < 0:
            ax2.set_xlim(-sf.start + 0.2, -sf.end - 0.2);
            ax2.set_xticks(np.arange(-sf.start, -sf.end - 1, -interval))
        else:
            ax2.set_xlim(sf.start - 0.2, sf.end + 0.2);
            ax2.set_xticks(np.arange(sf.start, sf.end + 1, interval))
    else:
        if sf.start < 0:
            ax2.set_xlim(-sf.start + 0.2, -sf.end - 0.2);
        else:
            ax2.set_xlim(sf.start - 0.2, sf.end + 0.2);
        ax2.xaxis.set_major_locator(MaxNLocator(integer=True))
    if savefig_path:
        plt.savefig(savefig_path)



import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
from scipy.fft import fft, ifft, fftfreq
from scipy.interpolate import interp1d
from functools import partial
from jax import jit
import jax.numpy as jnp
# import warnings
# warnings.filterwarnings("ignore")


class Resampler:

    def load_data(self, x, y, yerr):
        self.x = x
        self.y = y
        self.yerr = yerr
        self.start = np.nanmin(self.x)
        self.end = np.nanmax(self.x)
        self.annual = jnp.arange(self.start, self.end + 1)

    def super_gaussian(self, x, sz, order=30):
        return np.exp(-(x / sz) ** order)

    def bandpass_super_gaussian(self, signal, l, sz, order=30):
        if np.ndim(signal) == 1:
            signal = signal.reshape(1, -1)
        n = signal.shape[1]
        freq = fftfreq(n, d=1)
        mask = self.super_gaussian(np.abs(freq) - l, sz, order=order).reshape(1, -1)
        new_signal = ifft(fft(signal) * mask)
        return np.real(np.squeeze(new_signal))

    def signal_resample(self, size=10000, l=1./9.677, sz=0.05, order=30, samples=None):
        if not np.all(samples):
            samples = np.random.multivariate_normal(self.y, np.diag(self.yerr ** 2), size=size)
        f = interp1d(self.x, samples, kind="cubic")
        interpolation = f(self.annual)
        new_sig = self.bandpass_super_gaussian(interpolation, l, sz, order=order)
        return new_sig

